# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
from typing import Literal, TypeAlias

DefLocationDict: dict[str, int] = {
    "Location %u" % i: i for i in range(1, 256) if i != 64
}
DefLocationDict["Anywhere"] = 64

DefSwitchDict: dict[str, int] = {"Switch %d" % (i + 1): i for i in range(256)}

# ======================================

DefUnitDict: dict[str, int] = {
    "Terran Marine": 0,
    "Terran Ghost": 1,
    "Terran Vulture": 2,
    "Terran Goliath": 3,
    "Goliath Turret": 4,
    "Terran Siege Tank (Tank Mode)": 5,
    "Siege Tank Turret (Tank Mode)": 6,
    "Tank Turret type   1": 6,
    "Terran SCV": 7,
    "Terran Wraith": 8,
    "Terran Science Vessel": 9,
    "Gui Montag": 10,
    "Gui Montag (Firebat)": 10,
    "Terran Dropship": 11,
    "Terran Battlecruiser": 12,
    "Spider Mine": 13,
    "Vulture Spider Mine": 13,
    "Nuclear Missile": 14,
    "Terran Civilian": 15,
    "Sarah Kerrigan": 16,
    "Sarah Kerrigan (Ghost)": 16,
    "Alan Schezar": 17,
    "Alan Schezar (Goliath)": 17,
    "Alan Schezar Turret": 18,
    "Alan Turret": 18,
    "Jim Raynor (Vulture)": 19,
    "Jim Raynor (Marine)": 20,
    "Tom Kazansky": 21,
    "Tom Kazansky (Wraith)": 21,
    "Magellan": 22,
    "Magellan (Science Vessel)": 22,
    "Edmund Duke (Tank Mode)": 23,
    "Edmund Duke Turret (Tank Mode)": 24,
    "Duke Turret type   1": 24,
    "Edmund Duke (Siege Mode)": 25,
    "Edmund Duke Turret (Siege Mode)": 26,
    "Duke Turret type   2": 26,
    "Arcturus Mengsk": 27,
    "Arcturus Mengsk (Battlecruiser)": 27,
    "Hyperion": 28,
    "Hyperion (Battlecruiser)": 28,
    "Norad II (Battlecruiser)": 29,
    "Terran Siege Tank (Siege Mode)": 30,
    "Siege Tank Turret (Siege Mode)": 31,
    "Tank Turret type   2": 31,
    "Terran Firebat": 32,
    "Scanner Sweep": 33,
    "Terran Medic": 34,
    "Zerg Larva": 35,
    "Zerg Egg": 36,
    "Zerg Zergling": 37,
    "Zerg Hydralisk": 38,
    "Zerg Ultralisk": 39,
    "Zerg Broodling": 40,
    "Zerg Drone": 41,
    "Zerg Overlord": 42,
    "Zerg Mutalisk": 43,
    "Zerg Guardian": 44,
    "Zerg Queen": 45,
    "Zerg Defiler": 46,
    "Zerg Scourge": 47,
    "Torrasque": 48,
    "Torrasque (Ultralisk)": 48,
    "Matriarch": 49,
    "Matriarch (Queen)": 49,
    "Infested Terran": 50,
    "Infested Kerrigan": 51,
    "Infested Kerrigan (Infested Terran)": 51,
    "Unclean One": 52,
    "Unclean One (Defiler)": 52,
    "Hunter Killer": 53,
    "Hunter Killer (Hydralisk)": 53,
    "Devouring One": 54,
    "Devouring One (Zergling)": 54,
    "Kukulza (Mutalisk)": 55,
    "Kukulza (Guardian)": 56,
    "Yggdrasill": 57,
    "Yggdrasill (Overlord)": 57,
    "Terran Valkyrie": 58,
    "Mutalisk Cocoon": 59,
    "Cocoon": 59,
    "Protoss Corsair": 60,
    "Protoss Dark Templar": 61,
    "Protoss Dark Templar (Unit)": 61,
    "Zerg Devourer": 62,
    "Protoss Dark Archon": 63,
    "Protoss Probe": 64,
    "Protoss Zealot": 65,
    "Protoss Dragoon": 66,
    "Protoss High Templar": 67,
    "Protoss Archon": 68,
    "Protoss Shuttle": 69,
    "Protoss Scout": 70,
    "Protoss Arbiter": 71,
    "Protoss Carrier": 72,
    "Protoss Interceptor": 73,
    "Dark Templar": 74,
    "Dark Templar (Hero)": 74,
    "Protoss Dark Templar (Hero)": 74,
    "Zeratul": 75,
    "Zeratul (Dark Templar)": 75,
    "Tassadar/Zeratul": 76,
    "Tassadar/Zeratul (Archon)": 76,
    "Fenix (Zealot)": 77,
    "Fenix (Dragoon)": 78,
    "Tassadar": 79,
    "Tassadar (Templar)": 79,
    "Mojo": 80,
    "Mojo (Scout)": 80,
    "Warbringer": 81,
    "Warbringer (Reaver)": 81,
    "Gantrithor": 82,
    "Gantrithor (Carrier)": 82,
    "Protoss Reaver": 83,
    "Protoss Observer": 84,
    "Protoss Scarab": 85,
    "Danimoth": 86,
    "Danimoth (Arbiter)": 86,
    "Aldaris": 87,
    "Aldaris (Templar)": 87,
    "Artanis": 88,
    "Artanis (Scout)": 88,
    "Rhynadon": 89,
    "Rhynadon (Badlands)": 89,
    "Rhynadon (Badlands Critter)": 89,
    "Bengalaas": 90,
    "Bengalaas (Jungle)": 90,
    "Bengalaas (Jungle Critter)": 90,
    "Cargo Ship": 91,
    "Cargo Ship (Unused)": 91,
    "Unused type   1": 91,
    "Unused type   2": 92,
    "Mercenary Gunship": 92,
    "Mercenary Gunship (Unused)": 92,
    "Scantid": 93,
    "Scantid (Desert)": 93,
    "Scantid (Desert Critter)": 93,
    "Kakaru": 94,
    "Kakaru (Twilight)": 94,
    "Kakaru (Twilight Critter)": 94,
    "Ragnasaur": 95,
    "Ragnasaur (Ashworld)": 95,
    "Ragnasaur (Ashworld Critter)": 95,
    "Ursadon": 96,
    "Ursadon (Ice World)": 96,
    "Ursadon (Ice World Critter)": 96,
    "Lurker Egg": 97,
    "Zerg Lurker Egg": 97,
    "Raszagal": 98,
    "Raszagal (Corsair)": 98,
    "Samir Duran": 99,
    "Samir Duran (Ghost)": 99,
    "Alexei Stukov": 100,
    "Alexei Stukov (Ghost)": 100,
    "Map Revealer": 101,
    "Gerard DuGalle": 102,
    "Gerard DuGalle (Ghost)": 102,
    "Gerard DuGalle (BattleCruiser)": 102,
    "Zerg Lurker": 103,
    "Infested Duran": 104,
    "Infested Duran (Infested Terran)": 104,
    "Disruption Web": 105,
    "Disruption Field": 105,
    "Terran Command Center": 106,
    "Terran Comsat Station": 107,
    "Terran Nuclear Silo": 108,
    "Terran Supply Depot": 109,
    "Terran Refinery": 110,
    "Terran Barracks": 111,
    "Terran Academy": 112,
    "Terran Factory": 113,
    "Terran Starport": 114,
    "Terran Control Tower": 115,
    "Terran Science Facility": 116,
    "Terran Covert Ops": 117,
    "Terran Physics Lab": 118,
    "Starbase": 119,
    "Starbase (Unused)": 119,
    "Unused Terran Bldg type   1": 119,
    "Terran Machine Shop": 120,
    "Repair Bay": 121,
    "Repair Bay (Unused)": 121,
    "Unused Terran Bldg type   2": 121,
    "Terran Engineering Bay": 122,
    "Terran Armory": 123,
    "Terran Missile Turret": 124,
    "Terran Bunker": 125,
    "Norad II (Crashed)": 126,
    "Norad II (Crashed Battlecruiser)": 126,
    "Ion Cannon": 127,
    "Uraj Crystal": 128,
    "Khalis Crystal": 129,
    "Infested Command Center": 130,
    "Zerg Hatchery": 131,
    "Zerg Lair": 132,
    "Zerg Hive": 133,
    "Zerg Nydus Canal": 134,
    "Zerg Hydralisk Den": 135,
    "Zerg Defiler Mound": 136,
    "Zerg Greater Spire": 137,
    "Zerg Queen's Nest": 138,
    "Zerg Evolution Chamber": 139,
    "Zerg Ultralisk Cavern": 140,
    "Zerg Spire": 141,
    "Zerg Spawning Pool": 142,
    "Zerg Creep Colony": 143,
    "Zerg Spore Colony": 144,
    "Unused Zerg Bldg": 145,
    "Unused Zerg Building1": 145,
    "Zerg Sunken Colony": 146,
    "Zerg Overmind (With Shell)": 147,
    "Zerg Overmind": 148,
    "Zerg Extractor": 149,
    "Mature Chrysalis": 150,
    "Zerg Cerebrate": 151,
    "Zerg Cerebrate Daggoth": 152,
    "Unused Zerg Building2": 153,
    "Protoss Nexus": 154,
    "Protoss Robotics Facility": 155,
    "Protoss Pylon": 156,
    "Protoss Assimilator": 157,
    "Protoss Unused type   1": 158,
    "Unused Protoss Building1": 158,
    "Protoss Observatory": 159,
    "Protoss Gateway": 160,
    "Protoss Unused type   2": 161,
    "Unused Protoss Building2": 161,
    "Protoss Photon Cannon": 162,
    "Protoss Citadel of Adun": 163,
    "Protoss Cybernetics Core": 164,
    "Protoss Templar Archives": 165,
    "Protoss Forge": 166,
    "Protoss Stargate": 167,
    "Stasis Cell/Prison": 168,
    "Protoss Fleet Beacon": 169,
    "Protoss Arbiter Tribunal": 170,
    "Protoss Robotics Support Bay": 171,
    "Protoss Shield Battery": 172,
    "Khaydarin Crystal Formation": 173,
    "Protoss Temple": 174,
    "Xel'Naga Temple": 175,
    "Mineral Field (Type 1)": 176,
    "Mineral Field (Type 2)": 177,
    "Mineral Field (Type 3)": 178,
    "Cave": 179,
    "Cave (Unused)": 179,
    "Cave-in": 180,
    "Cave-in (Unused)": 180,
    "Cantina": 181,
    "Cantina (Unused)": 181,
    "Mining Platform": 182,
    "Mining Platform (Unused)": 182,
    "Independent Command Center": 183,
    "Independent Command Center (Unused)": 183,
    "Independent Starport": 184,
    "Independent Starport (Unused)": 184,
    "Jump Gate": 185,
    "Independent Jump Gate (Unused)": 185,
    "Ruins": 186,
    "Ruins (Unused)": 186,
    "Kyadarin Crystal Formation": 187,
    "Khaydarin Crystal Formation (Unused)": 187,
    "Vespene Geyser": 188,
    "Warp Gate": 189,
    "Psi Disrupter": 190,
    "Zerg Marker": 191,
    "Terran Marker": 192,
    "Protoss Marker": 193,
    "Zerg Beacon": 194,
    "Terran Beacon": 195,
    "Protoss Beacon": 196,
    "Zerg Flag Beacon": 197,
    "Terran Flag Beacon": 198,
    "Protoss Flag Beacon": 199,
    "Power Generator": 200,
    "Overmind Cocoon": 201,
    "Dark Swarm": 202,
    "Floor Missile Trap": 203,
    "Floor Hatch (Unused)": 204,
    "Floor Hatch (UNUSED)": 204,
    "Left Upper Level Door": 205,
    "Right Upper Level Door": 206,
    "Left Pit Door": 207,
    "Right Pit Door": 208,
    "Floor Gun Trap": 209,
    "Left Wall Missile Trap": 210,
    "Left Wall Flame Trap": 211,
    "Right Wall Missile Trap": 212,
    "Right Wall Flame Trap": 213,
    "Start Location": 214,
    "Flag": 215,
    "Young Chrysalis": 216,
    "Psi Emitter": 217,
    "Data Disc": 218,
    "Khaydarin Crystal": 219,
    "Mineral Chunk (Type 1)": 220,
    "Mineral Chunk (Type 2)": 221,
    "Mineral Cluster Type 1": 220,
    "Mineral Cluster Type 2": 221,
    "Vespene Orb (Protoss Type 1)": 222,
    "Vespene Orb (Protoss Type 2)": 223,
    "Protoss Vespene Gas Orb Type 1": 222,
    "Protoss Vespene Gas Orb Type 2": 223,
    "Vespene Sac (Zerg Type 1)": 224,
    "Vespene Sac (Zerg Type 2)": 225,
    "Zerg Vespene Gas Sac Type 1": 224,
    "Zerg Vespene Gas Sac Type 2": 225,
    "Vespene Tank (Terran Type 1)": 226,
    "Vespene Tank (Terran Type 2)": 227,
    "Terran Vespene Gas Tank Type 1": 226,
    "Terran Vespene Gas Tank Type 2": 227,
    "Unused unit 228": 228,
    "Unused228": 228,
    "None": 228,
    "Any unit": 229,
    "(any unit)": 229,
    "Men": 230,
    "(men)": 230,
    "Buildings": 231,
    "(buildings)": 231,
    "Factories": 232,
    "(factories)": 232,
}

DefaultUnit: TypeAlias = Literal[
    "Terran Marine",
    "Terran Ghost",
    "Terran Vulture",
    "Terran Goliath",
    "Goliath Turret",
    "Terran Siege Tank (Tank Mode)",
    "Siege Tank Turret (Tank Mode)",
    "Terran SCV",
    "Terran Wraith",
    "Terran Science Vessel",
    "Gui Montag",
    "Terran Dropship",
    "Terran Battlecruiser",
    "Vulture Spider Mine",
    "Nuclear Missile",
    "Terran Civilian",
    "Sarah Kerrigan",
    "Alan Schezar",
    "Alan Schezar Turret",
    "Jim Raynor (Vulture)",
    "Jim Raynor (Marine)",
    "Tom Kazansky",
    "Magellan",
    "Edmund Duke (Tank Mode)",
    "Edmund Duke Turret (Tank Mode)",
    "Edmund Duke (Siege Mode)",
    "Edmund Duke Turret (Siege Mode)",
    "Arcturus Mengsk",
    "Hyperion",
    "Norad II (Battlecruiser)",
    "Terran Siege Tank (Siege Mode)",
    "Siege Tank Turret (Siege Mode)",
    "Terran Firebat",
    "Scanner Sweep",
    "Terran Medic",
    "Zerg Larva",
    "Zerg Egg",
    "Zerg Zergling",
    "Zerg Hydralisk",
    "Zerg Ultralisk",
    "Zerg Broodling",
    "Zerg Drone",
    "Zerg Overlord",
    "Zerg Mutalisk",
    "Zerg Guardian",
    "Zerg Queen",
    "Zerg Defiler",
    "Zerg Scourge",
    "Torrasque",
    "Matriarch",
    "Infested Terran",
    "Infested Kerrigan",
    "Unclean One",
    "Hunter Killer",
    "Devouring One",
    "Kukulza (Mutalisk)",
    "Kukulza (Guardian)",
    "Yggdrasill",
    "Terran Valkyrie",
    "Cocoon",
    "Protoss Corsair",
    "Protoss Dark Templar",
    "Zerg Devourer",
    "Protoss Dark Archon",
    "Protoss Probe",
    "Protoss Zealot",
    "Protoss Dragoon",
    "Protoss High Templar",
    "Protoss Archon",
    "Protoss Shuttle",
    "Protoss Scout",
    "Protoss Arbiter",
    "Protoss Carrier",
    "Protoss Interceptor",
    "Dark Templar (Hero)",
    "Zeratul",
    "Tassadar/Zeratul",
    "Fenix (Zealot)",
    "Fenix (Dragoon)",
    "Tassadar",
    "Mojo"
    "Warbringer",
    "Gantrithor",
    "Protoss Reaver",
    "Protoss Observer",
    "Protoss Scarab",
    "Danimoth",
    "Aldaris",
    "Artanis",
    "Rhynadon",
    "Bengalaas",
    "Cargo Ship",
    "Mercenary Gunship",
    "Scantid",
    "Kakaru",
    "Ragnasaur",
    "Ursadon",
    "Lurker Egg",
    "Zerg Lurker Egg",
    "Raszagal",
    "Samir Duran",
    "Alexei Stukov",
    "Map Revealer",
    "Gerard DuGalle",
    "Zerg Lurker",
    "Infested Duran",
    "Disruption Web",
    "Terran Command Center",
    "Terran Comsat Station",
    "Terran Nuclear Silo",
    "Terran Supply Depot",
    "Terran Refinery",
    "Terran Barracks",
    "Terran Academy",
    "Terran Factory",
    "Terran Starport",
    "Terran Control Tower",
    "Terran Science Facility",
    "Terran Covert Ops",
    "Terran Physics Lab",
    "Starbase",
    "Terran Machine Shop",
    "Repair Bay",
    "Terran Engineering Bay",
    "Terran Armory",
    "Terran Missile Turret",
    "Terran Bunker",
    "Norad II (Crashed)",
    "Ion Cannon",
    "Uraj Crystal",
    "Khalis Crystal",
    "Infested Command Center",
    "Zerg Hatchery",
    "Zerg Lair",
    "Zerg Hive",
    "Zerg Nydus Canal",
    "Zerg Hydralisk Den",
    "Zerg Defiler Mound",
    "Zerg Greater Spire",
    "Zerg Queen's Nest",
    "Zerg Evolution Chamber",
    "Zerg Ultralisk Cavern",
    "Zerg Spire",
    "Zerg Spawning Pool",
    "Zerg Creep Colony",
    "Zerg Spore Colony",
    "Unused Zerg Building1",
    "Zerg Sunken Colony",
    "Zerg Overmind (With Shell)",
    "Zerg Overmind",
    "Zerg Extractor",
    "Mature Chrysalis",
    "Zerg Cerebrate",
    "Zerg Cerebrate Daggoth",
    "Unused Zerg Building2",
    "Protoss Nexus",
    "Protoss Robotics Facility",
    "Protoss Pylon",
    "Protoss Assimilator",
    "Unused Protoss Building1",
    "Protoss Observatory",
    "Protoss Gateway",
    "Unused Protoss Building2",
    "Protoss Photon Cannon",
    "Protoss Citadel of Adun",
    "Protoss Cybernetics Core",
    "Protoss Templar Archives",
    "Protoss Forge",
    "Protoss Stargate",
    "Stasis Cell/Prison",
    "Protoss Fleet Beacon",
    "Protoss Arbiter Tribunal",
    "Protoss Robotics Support Bay",
    "Protoss Shield Battery",
    "Khaydarin Crystal Formation",
    "Protoss Temple",
    "Xel'Naga Temple",
    "Mineral Field (Type 1)",
    "Mineral Field (Type 2)",
    "Mineral Field (Type 3)",
    "Cave",
    "Cave-in",
    "Cantina",
    "Mining Platform",
    "Independent Command Center",
    "Independent Starport",
    "Jump Gate",
    "Ruins",
    "Khaydarin Crystal Formation (Unused)",
    "Vespene Geyser",
    "Warp Gate",
    "Psi Disrupter",
    "Zerg Marker",
    "Terran Marker",
    "Protoss Marker",
    "Zerg Beacon",
    "Terran Beacon",
    "Protoss Beacon",
    "Zerg Flag Beacon",
    "Terran Flag Beacon",
    "Protoss Flag Beacon",
    "Power Generator",
    "Overmind Cocoon",
    "Dark Swarm",
    "Floor Missile Trap",
    "Floor Hatch (Unused)",
    "Left Upper Level Door",
    "Right Upper Level Door",
    "Left Pit Door",
    "Right Pit Door",
    "Floor Gun Trap",
    "Left Wall Missile Trap",
    "Left Wall Flame Trap",
    "Right Wall Missile Trap",
    "Right Wall Flame Trap",
    "Start Location",
    "Flag",
    "Young Chrysalis",
    "Psi Emitter",
    "Data Disc",
    "Khaydarin Crystal",
    "Mineral Chunk (Type 1)",
    "Mineral Chunk (Type 2)",
    "Vespene Orb (Protoss Type 1)",
    "Vespene Orb (Protoss Type 2)",
    "Vespene Sac (Zerg Type 1)",
    "Vespene Sac (Zerg Type 2)",
    "Vespene Tank (Terran Type 1)",
    "Vespene Tank (Terran Type 2)",
    "None",
    "(any unit)",
    "(men)",
    "(buildings)",
    "(factories)",
]

# ======================================

# Data from http://cafe.daum.net/rpgguild/6cWR/158
# Original data from ScAIEdit
DefAIScriptDict: dict[bytes, bytes] = {
    # Custom AI Scripts
    b"Terran Custom Level": b"TMCu",
    b"Zerg Custom Level": b"ZMCu",
    b"Protoss Custom Level": b"PMCu",
    b"Terran Expansion Custom Level": b"TMCx",
    b"Zerg Expansion Custom Level": b"ZMCx",
    b"Protoss Expansion Custom Level": b"PMCx",
    b"Terran Campaign Easy": b"TLOf",
    b"Terran Campaign Medium": b"TMED",
    b"Terran Campaign Difficult": b"THIf",
    b"Terran Campaign Insane": b"TSUP",
    b"Terran Campaign Area Town": b"TARE",
    b"Zerg Campaign Easy": b"ZLOf",
    b"Zerg Campaign Medium": b"ZMED",
    b"Zerg Campaign Difficult": b"ZHIf",
    b"Zerg Campaign Insane": b"ZSUP",
    b"Zerg Campaign Area Town": b"ZARE",
    b"Protoss Campaign Easy": b"PLOf",
    b"Protoss Campaign Medium": b"PMED",
    b"Protoss Campaign Difficult": b"PHIf",
    b"Protoss Campaign Insane": b"PSUP",
    b"Protoss Campaign Area Town": b"PARE",
    b"Expansion Terran Campaign Easy": b"TLOx",
    b"Expansion Terran Campaign Medium": b"TMEx",
    b"Expansion Terran Campaign Difficult": b"THIx",
    b"Expansion Terran Campaign Insane": b"TSUx",
    b"Expansion Terran Campaign Area Town": b"TARx",
    b"Expansion Zerg Campaign Easy": b"ZLOx",
    b"Expansion Zerg Campaign Medium": b"ZMEx",
    b"Expansion Zerg Campaign Difficult": b"ZHIx",
    b"Expansion Zerg Campaign Insane": b"ZSUx",
    b"Expansion Zerg Campaign Area Town": b"ZARx",
    b"Expansion Protoss Campaign Easy": b"PLOx",
    b"Expansion Protoss Campaign Medium": b"PMEx",
    b"Expansion Protoss Campaign Difficult": b"PHIx",
    b"Expansion Protoss Campaign Insane": b"PSUx",
    b"Expansion Protoss Campaign Area Town": b"PARx",
    b"Send All Units on Strategic Suicide Missions": b"Suic",
    b"Send All Units on Random Suicide Missions": b"SuiR",
    b"Switch Computer Player to Rescue Passive": b"Rscu",
    b"Turn ON Shared Vision for Player 1": b"+Vi0",
    b"Turn ON Shared Vision for Player 2": b"+Vi1",
    b"Turn ON Shared Vision for Player 3": b"+Vi2",
    b"Turn ON Shared Vision for Player 4": b"+Vi3",
    b"Turn ON Shared Vision for Player 5": b"+Vi4",
    b"Turn ON Shared Vision for Player 6": b"+Vi5",
    b"Turn ON Shared Vision for Player 7": b"+Vi6",
    b"Turn ON Shared Vision for Player 8": b"+Vi7",
    b"Turn OFF Shared Vision for Player 1": b"-Vi0",
    b"Turn OFF Shared Vision for Player 2": b"-Vi1",
    b"Turn OFF Shared Vision for Player 3": b"-Vi2",
    b"Turn OFF Shared Vision for Player 4": b"-Vi3",
    b"Turn OFF Shared Vision for Player 5": b"-Vi4",
    b"Turn OFF Shared Vision for Player 6": b"-Vi5",
    b"Turn OFF Shared Vision for Player 7": b"-Vi6",
    b"Turn OFF Shared Vision for Player 8": b"-Vi7",
    b"Move Dark Templars to Region": b"MvTe",
    b"Clear Previous Combat Data": b"ClrC",
    b"Set Player to Enemy": b"Enmy",
    b"Set Player to Ally": b"Ally",
    b"Value This Area Higher": b"VluA",
    b"Enter Closest Bunker": b"EnBk",
    b"Set Generic Command Target": b"StTg",
    b"Make These Units Patrol": b"StPt",
    b"Enter Transport": b"EnTr",
    b"Exit Transport": b"ExTr",
    b"AI Nuke Here": b"NuHe",
    b"AI Harass Here": b"HaHe",
    b"Set Unit Order To: Junk Yard Dog": b"JYDg",
    b"Disruption Web Here": b"DWHe",
    b"Recall Here": b"ReHe",
    # StarCraft AI Scripts
    b"Terran 3 - Zerg Town": b"Ter3",
    b"Terran 5 - Terran Main Town": b"Ter5",
    b"Terran 5 - Terran Harvest Town": b"Te5H",
    b"Terran 6 - Air Attack Zerg": b"Ter6",
    b"Terran 6 - Ground Attack Zerg": b"Te6b",
    b"Terran 6 - Zerg Support Town": b"Te6c",
    b"Terran 7 - Bottom Zerg Town": b"Ter7",
    b"Terran 7 - Right Zerg Town": b"Te7s",
    b"Terran 7 - Middle Zerg Town": b"Te7m",
    b"Terran 8 - Confederate Town": b"Ter8",
    b"Terran 9 - Light Attack": b"Tr9L",
    b"Terran 9 - Heavy Attack": b"Tr9H",
    b"Terran 10 - Confederate Towns": b"Te10",
    b"Terran 11 - Zerg Town": b"T11z",
    b"Terran 11 - Lower Protoss Town": b"T11a",
    b"Terran 11 - Upper Protoss Town": b"T11b",
    b"Terran 12 - Nuke Town": b"T12N",
    b"Terran 12 - Phoenix Town": b"T12P",
    b"Terran 12 - Tank Town": b"T12T",
    b"Terran 1 - Electronic Distribution": b"TED1",
    b"Terran 2 - Electronic Distribution": b"TED2",
    b"Terran 3 - Electronic Distribution": b"TED3",
    b"Terran 1 - Shareware": b"TSW1",
    b"Terran 2 - Shareware": b"TSW2",
    b"Terran 3 - Shareware": b"TSW3",
    b"Terran 4 - Shareware": b"TSW4",
    b"Terran 5 - Shareware": b"TSW5",
    b"Zerg 1 - Terran Town": b"Zer1",
    b"Zerg 2 - Protoss Town": b"Zer2",
    b"Zerg 3 - Terran Town": b"Zer3",
    b"Zerg 4 - Right Terran Town": b"Zer4",
    b"Zerg 4 - Lower Terran Town": b"Ze4S",
    b"Zerg 6 - Protoss Town": b"Zer6",
    b"Zerg 7 - Air Town": b"Zr7a",
    b"Zerg 7 - Ground Town": b"Zr7g",
    b"Zerg 7 - Support Town": b"Zr7s",
    b"Zerg 8 - Scout Town": b"Zer8",
    b"Zerg 8 - Templar Town": b"Ze8T",
    b"Zerg 9 - Teal Protoss": b"Zer9",
    b"Zerg 9 - Left Yellow Protoss": b"Z9ly",
    b"Zerg 9 - Right Yellow Protoss": b"Z9ry",
    b"Zerg 9 - Left Orange Protoss": b"Z9lo",
    b"Zerg 9 - Right Orange Protoss": b"Z9ro",
    b"Zerg 10 - Left Teal (Attack": b"Z10a",
    b"Zerg 10 - Right Teal (Support": b"Z10b",
    b"Zerg 10 - Left Yellow (Support": b"Z10c",
    b"Zerg 10 - Right Yellow (Attack": b"Z10d",
    b"Zerg 10 - Red Protoss": b"Z10e",
    b"Protoss 1 - Zerg Town": b"Pro1",
    b"Protoss 2 - Zerg Town": b"Pro2",
    b"Protoss 3 - Air Zerg Town": b"Pr3R",
    b"Protoss 3 - Ground Zerg Town": b"Pr3G",
    b"Protoss 4 - Zerg Town": b"Pro4",
    b"Protoss 5 - Zerg Town Island": b"Pr5I",
    b"Protoss 5 - Zerg Town Base": b"Pr5B",
    b"Protoss 7 - Left Protoss Town": b"Pro7",
    b"Protoss 7 - Right Protoss Town": b"Pr7B",
    b"Protoss 7 - Shrine Protoss": b"Pr7S",
    b"Protoss 8 - Left Protoss Town": b"Pro8",
    b"Protoss 8 - Right Protoss Town": b"Pr8B",
    b"Protoss 8 - Protoss Defenders": b"Pr8D",
    b"Protoss 9 - Ground Zerg": b"Pro9",
    b"Protoss 9 - Air Zerg": b"Pr9W",
    b"Protoss 9 - Spell Zerg": b"Pr9Y",
    b"Protoss 10 - Mini-Towns": b"Pr10",
    b"Protoss 10 - Mini-Town Master": b"P10C",
    b"Protoss 10 - Overmind Defenders": b"P10o",
    # Brood Wars AI Scripts
    b"Brood Wars Protoss 1 - Town A": b"PB1A",
    b"Brood Wars Protoss 1 - Town B": b"PB1B",
    b"Brood Wars Protoss 1 - Town C": b"PB1C",
    b"Brood Wars Protoss 1 - Town D": b"PB1D",
    b"Brood Wars Protoss 1 - Town E": b"PB1E",
    b"Brood Wars Protoss 1 - Town F": b"PB1F",
    b"Brood Wars Protoss 2 - Town A": b"PB2A",
    b"Brood Wars Protoss 2 - Town B": b"PB2B",
    b"Brood Wars Protoss 2 - Town C": b"PB2C",
    b"Brood Wars Protoss 2 - Town D": b"PB2D",
    b"Brood Wars Protoss 2 - Town E": b"PB2E",
    b"Brood Wars Protoss 2 - Town F": b"PB2F",
    b"Brood Wars Protoss 3 - Town A": b"PB3A",
    b"Brood Wars Protoss 3 - Town B": b"PB3B",
    b"Brood Wars Protoss 3 - Town C": b"PB3C",
    b"Brood Wars Protoss 3 - Town D": b"PB3D",
    b"Brood Wars Protoss 3 - Town E": b"PB3E",
    b"Brood Wars Protoss 3 - Town F": b"PB3F",
    b"Brood Wars Protoss 4 - Town A": b"PB4A",
    b"Brood Wars Protoss 4 - Town B": b"PB4B",
    b"Brood Wars Protoss 4 - Town C": b"PB4C",
    b"Brood Wars Protoss 4 - Town D": b"PB4D",
    b"Brood Wars Protoss 4 - Town E": b"PB4E",
    b"Brood Wars Protoss 4 - Town F": b"PB4F",
    b"Brood Wars Protoss 5 - Town A": b"PB5A",
    b"Brood Wars Protoss 5 - Town B": b"PB5B",
    b"Brood Wars Protoss 5 - Town C": b"PB5C",
    b"Brood Wars Protoss 5 - Town D": b"PB5D",
    b"Brood Wars Protoss 5 - Town E": b"PB5E",
    b"Brood Wars Protoss 5 - Town F": b"PB5F",
    b"Brood Wars Protoss 6 - Town A": b"PB6A",
    b"Brood Wars Protoss 6 - Town B": b"PB6B",
    b"Brood Wars Protoss 6 - Town C": b"PB6C",
    b"Brood Wars Protoss 6 - Town D": b"PB6D",
    b"Brood Wars Protoss 6 - Town E": b"PB6E",
    b"Brood Wars Protoss 6 - Town F": b"PB6F",
    b"Brood Wars Protoss 7 - Town A": b"PB7A",
    b"Brood Wars Protoss 7 - Town B": b"PB7B",
    b"Brood Wars Protoss 7 - Town C": b"PB7C",
    b"Brood Wars Protoss 7 - Town D": b"PB7D",
    b"Brood Wars Protoss 7 - Town E": b"PB7E",
    b"Brood Wars Protoss 7 - Town F": b"PB7F",
    b"Brood Wars Protoss 8 - Town A": b"PB8A",
    b"Brood Wars Protoss 8 - Town B": b"PB8B",
    b"Brood Wars Protoss 8 - Town C": b"PB8C",
    b"Brood Wars Protoss 8 - Town D": b"PB8D",
    b"Brood Wars Protoss 8 - Town E": b"PB8E",
    b"Brood Wars Protoss 8 - Town F": b"PB8F",
    b"Brood Wars Terran 1 - Town A": b"TB1A",
    b"Brood Wars Terran 1 - Town B": b"TB1B",
    b"Brood Wars Terran 1 - Town C": b"TB1C",
    b"Brood Wars Terran 1 - Town D": b"TB1D",
    b"Brood Wars Terran 1 - Town E": b"TB1E",
    b"Brood Wars Terran 1 - Town F": b"TB1F",
    b"Brood Wars Terran 2 - Town A": b"TB2A",
    b"Brood Wars Terran 2 - Town B": b"TB2B",
    b"Brood Wars Terran 2 - Town C": b"TB2C",
    b"Brood Wars Terran 2 - Town D": b"TB2D",
    b"Brood Wars Terran 2 - Town E": b"TB2E",
    b"Brood Wars Terran 2 - Town F": b"TB2F",
    b"Brood Wars Terran 3 - Town A": b"TB3A",
    b"Brood Wars Terran 3 - Town B": b"TB3B",
    b"Brood Wars Terran 3 - Town C": b"TB3C",
    b"Brood Wars Terran 3 - Town D": b"TB3D",
    b"Brood Wars Terran 3 - Town E": b"TB3E",
    b"Brood Wars Terran 3 - Town F": b"TB3F",
    b"Brood Wars Terran 4 - Town A": b"TB4A",
    b"Brood Wars Terran 4 - Town B": b"TB4B",
    b"Brood Wars Terran 4 - Town C": b"TB4C",
    b"Brood Wars Terran 4 - Town D": b"TB4D",
    b"Brood Wars Terran 4 - Town E": b"TB4E",
    b"Brood Wars Terran 4 - Town F": b"TB4F",
    b"Brood Wars Terran 5 - Town A": b"TB5A",
    b"Brood Wars Terran 5 - Town B": b"TB5B",
    b"Brood Wars Terran 5 - Town C": b"TB5C",
    b"Brood Wars Terran 5 - Town D": b"TB5D",
    b"Brood Wars Terran 5 - Town E": b"TB5E",
    b"Brood Wars Terran 5 - Town F": b"TB5F",
    b"Brood Wars Terran 6 - Town A": b"TB6A",
    b"Brood Wars Terran 6 - Town B": b"TB6B",
    b"Brood Wars Terran 6 - Town C": b"TB6C",
    b"Brood Wars Terran 6 - Town D": b"TB6D",
    b"Brood Wars Terran 6 - Town E": b"TB6E",
    b"Brood Wars Terran 6 - Town F": b"TB6F",
    b"Brood Wars Terran 7 - Town A": b"TB7A",
    b"Brood Wars Terran 7 - Town B": b"TB7B",
    b"Brood Wars Terran 7 - Town C": b"TB7C",
    b"Brood Wars Terran 7 - Town D": b"TB7D",
    b"Brood Wars Terran 7 - Town E": b"TB7E",
    b"Brood Wars Terran 7 - Town F": b"TB7F",
    b"Brood Wars Terran 8 - Town A": b"TB8A",
    b"Brood Wars Terran 8 - Town B": b"TB8B",
    b"Brood Wars Terran 8 - Town C": b"TB8C",
    b"Brood Wars Terran 8 - Town D": b"TB8D",
    b"Brood Wars Terran 8 - Town E": b"TB8E",
    b"Brood Wars Terran 8 - Town F": b"TB8F",
    b"Brood Wars Zerg 1 - Town A": b"ZB1A",
    b"Brood Wars Zerg 1 - Town B": b"ZB1B",
    b"Brood Wars Zerg 1 - Town C": b"ZB1C",
    b"Brood Wars Zerg 1 - Town D": b"ZB1D",
    b"Brood Wars Zerg 1 - Town E": b"ZB1E",
    b"Brood Wars Zerg 1 - Town F": b"ZB1F",
    b"Brood Wars Zerg 2 - Town A": b"ZB2A",
    b"Brood Wars Zerg 2 - Town B": b"ZB2B",
    b"Brood Wars Zerg 2 - Town C": b"ZB2C",
    b"Brood Wars Zerg 2 - Town D": b"ZB2D",
    b"Brood Wars Zerg 2 - Town E": b"ZB2E",
    b"Brood Wars Zerg 2 - Town F": b"ZB2F",
    b"Brood Wars Zerg 3 - Town A": b"ZB3A",
    b"Brood Wars Zerg 3 - Town B": b"ZB3B",
    b"Brood Wars Zerg 3 - Town C": b"ZB3C",
    b"Brood Wars Zerg 3 - Town D": b"ZB3D",
    b"Brood Wars Zerg 3 - Town E": b"ZB3E",
    b"Brood Wars Zerg 3 - Town F": b"ZB3F",
    b"Brood Wars Zerg 4 - Town A": b"ZB4A",
    b"Brood Wars Zerg 4 - Town B": b"ZB4B",
    b"Brood Wars Zerg 4 - Town C": b"ZB4C",
    b"Brood Wars Zerg 4 - Town D": b"ZB4D",
    b"Brood Wars Zerg 4 - Town E": b"ZB4E",
    b"Brood Wars Zerg 4 - Town F": b"ZB4F",
    b"Brood Wars Zerg 5 - Town A": b"ZB5A",
    b"Brood Wars Zerg 5 - Town B": b"ZB5B",
    b"Brood Wars Zerg 5 - Town C": b"ZB5C",
    b"Brood Wars Zerg 5 - Town D": b"ZB5D",
    b"Brood Wars Zerg 5 - Town E": b"ZB5E",
    b"Brood Wars Zerg 5 - Town F": b"ZB5F",
    b"Brood Wars Zerg 6 - Town A": b"ZB6A",
    b"Brood Wars Zerg 6 - Town B": b"ZB6B",
    b"Brood Wars Zerg 6 - Town C": b"ZB6C",
    b"Brood Wars Zerg 6 - Town D": b"ZB6D",
    b"Brood Wars Zerg 6 - Town E": b"ZB6E",
    b"Brood Wars Zerg 6 - Town F": b"ZB6F",
    b"Brood Wars Zerg 7 - Town A": b"ZB7A",
    b"Brood Wars Zerg 7 - Town B": b"ZB7B",
    b"Brood Wars Zerg 7 - Town C": b"ZB7C",
    b"Brood Wars Zerg 7 - Town D": b"ZB7D",
    b"Brood Wars Zerg 7 - Town E": b"ZB7E",
    b"Brood Wars Zerg 7 - Town F": b"ZB7F",
    b"Brood Wars Zerg 8 - Town A": b"ZB8A",
    b"Brood Wars Zerg 8 - Town B": b"ZB8B",
    b"Brood Wars Zerg 8 - Town C": b"ZB8C",
    b"Brood Wars Zerg 8 - Town D": b"ZB8D",
    b"Brood Wars Zerg 8 - Town E": b"ZB8E",
    b"Brood Wars Zerg 8 - Town F": b"ZB8F",
    b"Brood Wars Zerg 9 - Town A": b"ZB9A",
    b"Brood Wars Zerg 9 - Town B": b"ZB9B",
    b"Brood Wars Zerg 9 - Town C": b"ZB9C",
    b"Brood Wars Zerg 9 - Town D": b"ZB9D",
    b"Brood Wars Zerg 9 - Town E": b"ZB9E",
    b"Brood Wars Zerg 9 - Town F": b"ZB9F",
    b"Brood Wars Zerg 10 - Town A": b"ZB0A",
    b"Brood Wars Zerg 10 - Town B": b"ZB0B",
    b"Brood Wars Zerg 10 - Town C": b"ZB0C",
    b"Brood Wars Zerg 10 - Town D": b"ZB0D",
    b"Brood Wars Zerg 10 - Town E": b"ZB0E",
    b"Brood Wars Zerg 10 - Town F": b"ZB0F",
}

DefaultAIScriptWithoutLocation: TypeAlias = Literal[
    "Send All Units on Strategic Suicide Missions",
    "Send All Units on Random Suicide Missions",
    "Switch Computer Player to Rescue Passive",
    "Turn ON Shared Vision for Player 1",
    "Turn ON Shared Vision for Player 2",
    "Turn ON Shared Vision for Player 3",
    "Turn ON Shared Vision for Player 4",
    "Turn ON Shared Vision for Player 5",
    "Turn ON Shared Vision for Player 6",
    "Turn ON Shared Vision for Player 7",
    "Turn ON Shared Vision for Player 8",
    "Turn OFF Shared Vision for Player 1",
    "Turn OFF Shared Vision for Player 2",
    "Turn OFF Shared Vision for Player 3",
    "Turn OFF Shared Vision for Player 4",
    "Turn OFF Shared Vision for Player 5",
    "Turn OFF Shared Vision for Player 6",
    "Turn OFF Shared Vision for Player 7",
    "Turn OFF Shared Vision for Player 8",
]

DefaultAIScriptAtLocation: TypeAlias = Literal[
    "Terran Custom Level",
    "Zerg Custom Level",
    "Protoss Custom Level",
    "Terran Expansion Custom Level",
    "Zerg Expansion Custom Level",
    "Protoss Expansion Custom Level",
    "Terran Campaign Easy",
    "Terran Campaign Medium",
    "Terran Campaign Difficult",
    "Terran Campaign Insane",
    "Terran Campaign Area Town",
    "Zerg Campaign Easy",
    "Zerg Campaign Medium",
    "Zerg Campaign Difficult",
    "Zerg Campaign Insane",
    "Zerg Campaign Area Town",
    "Protoss Campaign Easy",
    "Protoss Campaign Medium",
    "Protoss Campaign Difficult",
    "Protoss Campaign Insane",
    "Protoss Campaign Area Town",
    "Expansion Terran Campaign Easy",
    "Expansion Terran Campaign Medium",
    "Expansion Terran Campaign Difficult",
    "Expansion Terran Campaign Insane",
    "Expansion Terran Campaign Area Town",
    "Expansion Zerg Campaign Easy",
    "Expansion Zerg Campaign Medium",
    "Expansion Zerg Campaign Difficult",
    "Expansion Zerg Campaign Insane",
    "Expansion Zerg Campaign Area Town",
    "Expansion Protoss Campaign Easy",
    "Expansion Protoss Campaign Medium",
    "Expansion Protoss Campaign Difficult",
    "Expansion Protoss Campaign Insane",
    "Expansion Protoss Campaign Area Town",
    "Move Dark Templars to Region",
    "Clear Previous Combat Data",
    "Set Player to Enemy",
    "Set Player to Ally",
    "Value This Area Higher",
    "Enter Closest Bunker",
    "Set Generic Command Target",
    "Make These Units Patrol",
    "Enter Transport",
    "Exit Transport",
    "AI Nuke Here",
    "AI Harass Here",
    "Set Unit Order To: Junk Yard Dog",
    "Disruption Web Here",
    "Recall Here",
    "Terran 3 - Zerg Town",
    "Terran 5 - Terran Main Town",
    "Terran 5 - Terran Harvest Town",
    "Terran 6 - Air Attack Zerg",
    "Terran 6 - Ground Attack Zerg",
    "Terran 6 - Zerg Support Town",
    "Terran 7 - Bottom Zerg Town",
    "Terran 7 - Right Zerg Town",
    "Terran 7 - Middle Zerg Town",
    "Terran 8 - Confederate Town",
    "Terran 9 - Light Attack",
    "Terran 9 - Heavy Attack",
    "Terran 10 - Confederate Towns",
    "Terran 11 - Zerg Town",
    "Terran 11 - Lower Protoss Town",
    "Terran 11 - Upper Protoss Town",
    "Terran 12 - Nuke Town",
    "Terran 12 - Phoenix Town",
    "Terran 12 - Tank Town",
    "Terran 1 - Electronic Distribution",
    "Terran 2 - Electronic Distribution",
    "Terran 3 - Electronic Distribution",
    "Terran 1 - Shareware",
    "Terran 2 - Shareware",
    "Terran 3 - Shareware",
    "Terran 4 - Shareware",
    "Terran 5 - Shareware",
    "Zerg 1 - Terran Town",
    "Zerg 2 - Protoss Town",
    "Zerg 3 - Terran Town",
    "Zerg 4 - Right Terran Town",
    "Zerg 4 - Lower Terran Town",
    "Zerg 6 - Protoss Town",
    "Zerg 7 - Air Town",
    "Zerg 7 - Ground Town",
    "Zerg 7 - Support Town",
    "Zerg 8 - Scout Town",
    "Zerg 8 - Templar Town",
    "Zerg 9 - Teal Protoss",
    "Zerg 9 - Left Yellow Protoss",
    "Zerg 9 - Right Yellow Protoss",
    "Zerg 9 - Left Orange Protoss",
    "Zerg 9 - Right Orange Protoss",
    "Zerg 10 - Left Teal (Attack",
    "Zerg 10 - Right Teal (Support",
    "Zerg 10 - Left Yellow (Support",
    "Zerg 10 - Right Yellow (Attack",
    "Zerg 10 - Red Protoss",
    "Protoss 1 - Zerg Town",
    "Protoss 2 - Zerg Town",
    "Protoss 3 - Air Zerg Town",
    "Protoss 3 - Ground Zerg Town",
    "Protoss 4 - Zerg Town",
    "Protoss 5 - Zerg Town Island",
    "Protoss 5 - Zerg Town Base",
    "Protoss 7 - Left Protoss Town",
    "Protoss 7 - Right Protoss Town",
    "Protoss 7 - Shrine Protoss",
    "Protoss 8 - Left Protoss Town",
    "Protoss 8 - Right Protoss Town",
    "Protoss 8 - Protoss Defenders",
    "Protoss 9 - Ground Zerg",
    "Protoss 9 - Air Zerg",
    "Protoss 9 - Spell Zerg",
    "Protoss 10 - Mini-Towns",
    "Protoss 10 - Mini-Town Master",
    "Protoss 10 - Overmind Defenders",
    "Brood Wars Protoss 1 - Town A",
    "Brood Wars Protoss 1 - Town B",
    "Brood Wars Protoss 1 - Town C",
    "Brood Wars Protoss 1 - Town D",
    "Brood Wars Protoss 1 - Town E",
    "Brood Wars Protoss 1 - Town F",
    "Brood Wars Protoss 2 - Town A",
    "Brood Wars Protoss 2 - Town B",
    "Brood Wars Protoss 2 - Town C",
    "Brood Wars Protoss 2 - Town D",
    "Brood Wars Protoss 2 - Town E",
    "Brood Wars Protoss 2 - Town F",
    "Brood Wars Protoss 3 - Town A",
    "Brood Wars Protoss 3 - Town B",
    "Brood Wars Protoss 3 - Town C",
    "Brood Wars Protoss 3 - Town D",
    "Brood Wars Protoss 3 - Town E",
    "Brood Wars Protoss 3 - Town F",
    "Brood Wars Protoss 4 - Town A",
    "Brood Wars Protoss 4 - Town B",
    "Brood Wars Protoss 4 - Town C",
    "Brood Wars Protoss 4 - Town D",
    "Brood Wars Protoss 4 - Town E",
    "Brood Wars Protoss 4 - Town F",
    "Brood Wars Protoss 5 - Town A",
    "Brood Wars Protoss 5 - Town B",
    "Brood Wars Protoss 5 - Town C",
    "Brood Wars Protoss 5 - Town D",
    "Brood Wars Protoss 5 - Town E",
    "Brood Wars Protoss 5 - Town F",
    "Brood Wars Protoss 6 - Town A",
    "Brood Wars Protoss 6 - Town B",
    "Brood Wars Protoss 6 - Town C",
    "Brood Wars Protoss 6 - Town D",
    "Brood Wars Protoss 6 - Town E",
    "Brood Wars Protoss 6 - Town F",
    "Brood Wars Protoss 7 - Town A",
    "Brood Wars Protoss 7 - Town B",
    "Brood Wars Protoss 7 - Town C",
    "Brood Wars Protoss 7 - Town D",
    "Brood Wars Protoss 7 - Town E",
    "Brood Wars Protoss 7 - Town F",
    "Brood Wars Protoss 8 - Town A",
    "Brood Wars Protoss 8 - Town B",
    "Brood Wars Protoss 8 - Town C",
    "Brood Wars Protoss 8 - Town D",
    "Brood Wars Protoss 8 - Town E",
    "Brood Wars Protoss 8 - Town F",
    "Brood Wars Terran 1 - Town A",
    "Brood Wars Terran 1 - Town B",
    "Brood Wars Terran 1 - Town C",
    "Brood Wars Terran 1 - Town D",
    "Brood Wars Terran 1 - Town E",
    "Brood Wars Terran 1 - Town F",
    "Brood Wars Terran 2 - Town A",
    "Brood Wars Terran 2 - Town B",
    "Brood Wars Terran 2 - Town C",
    "Brood Wars Terran 2 - Town D",
    "Brood Wars Terran 2 - Town E",
    "Brood Wars Terran 2 - Town F",
    "Brood Wars Terran 3 - Town A",
    "Brood Wars Terran 3 - Town B",
    "Brood Wars Terran 3 - Town C",
    "Brood Wars Terran 3 - Town D",
    "Brood Wars Terran 3 - Town E",
    "Brood Wars Terran 3 - Town F",
    "Brood Wars Terran 4 - Town A",
    "Brood Wars Terran 4 - Town B",
    "Brood Wars Terran 4 - Town C",
    "Brood Wars Terran 4 - Town D",
    "Brood Wars Terran 4 - Town E",
    "Brood Wars Terran 4 - Town F",
    "Brood Wars Terran 5 - Town A",
    "Brood Wars Terran 5 - Town B",
    "Brood Wars Terran 5 - Town C",
    "Brood Wars Terran 5 - Town D",
    "Brood Wars Terran 5 - Town E",
    "Brood Wars Terran 5 - Town F",
    "Brood Wars Terran 6 - Town A",
    "Brood Wars Terran 6 - Town B",
    "Brood Wars Terran 6 - Town C",
    "Brood Wars Terran 6 - Town D",
    "Brood Wars Terran 6 - Town E",
    "Brood Wars Terran 6 - Town F",
    "Brood Wars Terran 7 - Town A",
    "Brood Wars Terran 7 - Town B",
    "Brood Wars Terran 7 - Town C",
    "Brood Wars Terran 7 - Town D",
    "Brood Wars Terran 7 - Town E",
    "Brood Wars Terran 7 - Town F",
    "Brood Wars Terran 8 - Town A",
    "Brood Wars Terran 8 - Town B",
    "Brood Wars Terran 8 - Town C",
    "Brood Wars Terran 8 - Town D",
    "Brood Wars Terran 8 - Town E",
    "Brood Wars Terran 8 - Town F",
    "Brood Wars Zerg 1 - Town A",
    "Brood Wars Zerg 1 - Town B",
    "Brood Wars Zerg 1 - Town C",
    "Brood Wars Zerg 1 - Town D",
    "Brood Wars Zerg 1 - Town E",
    "Brood Wars Zerg 1 - Town F",
    "Brood Wars Zerg 2 - Town A",
    "Brood Wars Zerg 2 - Town B",
    "Brood Wars Zerg 2 - Town C",
    "Brood Wars Zerg 2 - Town D",
    "Brood Wars Zerg 2 - Town E",
    "Brood Wars Zerg 2 - Town F",
    "Brood Wars Zerg 3 - Town A",
    "Brood Wars Zerg 3 - Town B",
    "Brood Wars Zerg 3 - Town C",
    "Brood Wars Zerg 3 - Town D",
    "Brood Wars Zerg 3 - Town E",
    "Brood Wars Zerg 3 - Town F",
    "Brood Wars Zerg 4 - Town A",
    "Brood Wars Zerg 4 - Town B",
    "Brood Wars Zerg 4 - Town C",
    "Brood Wars Zerg 4 - Town D",
    "Brood Wars Zerg 4 - Town E",
    "Brood Wars Zerg 4 - Town F",
    "Brood Wars Zerg 5 - Town A",
    "Brood Wars Zerg 5 - Town B",
    "Brood Wars Zerg 5 - Town C",
    "Brood Wars Zerg 5 - Town D",
    "Brood Wars Zerg 5 - Town E",
    "Brood Wars Zerg 5 - Town F",
    "Brood Wars Zerg 6 - Town A",
    "Brood Wars Zerg 6 - Town B",
    "Brood Wars Zerg 6 - Town C",
    "Brood Wars Zerg 6 - Town D",
    "Brood Wars Zerg 6 - Town E",
    "Brood Wars Zerg 6 - Town F",
    "Brood Wars Zerg 7 - Town A",
    "Brood Wars Zerg 7 - Town B",
    "Brood Wars Zerg 7 - Town C",
    "Brood Wars Zerg 7 - Town D",
    "Brood Wars Zerg 7 - Town E",
    "Brood Wars Zerg 7 - Town F",
    "Brood Wars Zerg 8 - Town A",
    "Brood Wars Zerg 8 - Town B",
    "Brood Wars Zerg 8 - Town C",
    "Brood Wars Zerg 8 - Town D",
    "Brood Wars Zerg 8 - Town E",
    "Brood Wars Zerg 8 - Town F",
    "Brood Wars Zerg 9 - Town A",
    "Brood Wars Zerg 9 - Town B",
    "Brood Wars Zerg 9 - Town C",
    "Brood Wars Zerg 9 - Town D",
    "Brood Wars Zerg 9 - Town E",
    "Brood Wars Zerg 9 - Town F",
    "Brood Wars Zerg 10 - Town A",
    "Brood Wars Zerg 10 - Town B",
    "Brood Wars Zerg 10 - Town C",
    "Brood Wars Zerg 10 - Town D",
    "Brood Wars Zerg 10 - Town E",
    "Brood Wars Zerg 10 - Town F",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import random

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut

CP = 13
cpcache = c.curpl.GetCPCache()
cpcond = c.curpl.cpcache_match_cond()


def f_setcurpl(cp, *, actions=[], set_modifier=True):
    if c.IsEUDVariable(cp):
        c.VProc(
            cp,
            actions + cp.QueueAssignTo(cpcache)
            if set_modifier
            else [cp.SetDest(cpcache)],
        )
        c.VProc(
            [cp, cpcache],
            [
                cp.SetDest(ut.EPD(cpcond) + 2),
                cpcache.SetDest(ut.EPD(0x6509B0)),
            ],
        )
    else:
        cs.DoActions(c.SetCurrentPlayer(cp))


def f_setcurpl2cpcache(v=[], actions=[]):
    v = ut.FlattenList(v)
    actions = ut.FlattenList(actions)
    trg = c.VProc([*v, cpcache], [*actions, cpcache.SetDest(ut.EPD(0x6509B0))])

    return trg


# This function initializes _curpl_checkcond, so should be called at least once
def _f_updatecpcache():
    from .s import SetMemoryC

    c.RawTrigger(actions=SetMemoryC(cpcache.getValueAddr(), c.SetTo, 0))
    for i in ut._rand_lst(range(32)):
        u = random.randint(234, 65535)
        c.RawTrigger(
            conditions=c.DeathsX(ut.EPD(0x6509B0) - 12 * u, c.AtLeast, 1, u, 2**i),
            actions=SetMemoryC(cpcache.getValueAddr(), c.Add, 2**i),
        )
    c.VProc(cpcache, cpcache.SetDest(ut.EPD(cpcond) + 2))
    f_setcurpl2cpcache()


@c.EUDFunc
def f_getcurpl():
    """Get current player value.

    eudplib internally caches the current player value, so this function uses
    that value if the value is valid. Otherwise, update the current player
    cache and return it.
    """
    if cs.EUDIfNot()(cpcond):
        _f_updatecpcache()
    cs.EUDEndIf()

    return cpcache  # its dest can vary


def f_addcurpl(cp):
    """Add current player value.

    eudplib internally caches the current player value,
    so this function add to that value.
    """
    if c.IsEUDVariable(cp):
        c.VProc(cp, cp.QueueAddTo(cpcache))
        c.VProc(cp, cp.SetDest(ut.EPD(cpcond) + 2))
        c.VProc(cp, cp.SetDest(ut.EPD(0x6509B0)))
    else:
        cs.DoActions(c.AddCurrentPlayer(cp))


c.PushTriggerScope()
f_getcurpl()  # Force initialize f_getcurpl
c.PopTriggerScope()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import struct
from collections.abc import Iterator, Sequence
from typing import Literal, TypeAlias, overload

from typing_extensions import Self

from ... import utils as ut
from ...bindings._rust import allocator as alc
from ...localize import _
from ..allocator import ConstExpr, IsConstExpr
from ..allocator.payload import ObjCollector
from ..eudobj import EUDObject
from .action import Action
from .condition import Condition
from .triggerscope import NextTrigger, _register_trigger

# Trigger counter thing

_trg_counter = 0


def GetTriggerCounter() -> int:  # noqa: N802
    return _trg_counter


# Aux


@overload
def Disabled(arg: Condition) -> Condition:
    ...


@overload
def Disabled(arg: Action) -> Action:
    ...


def Disabled(arg: Condition | Action) -> Condition | Action:  # noqa: N802
    """Disable condition or action"""
    arg.disable()
    return arg


Trigger: TypeAlias = ConstExpr | int | None
_Condition: TypeAlias = Condition | bool | Sequence[Condition | bool | Sequence]
_Action: TypeAlias = Action | Sequence[Action | Sequence]


class RawTrigger(EUDObject):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls)

    @overload
    def __init__(
        self,
        prevptr: Trigger = None,
        nextptr: Trigger = None,
        conditions: _Condition | None = None,
        actions: _Action | None = None,
        *,
        preserved: bool = True,
        trigSection: None = None,  # noqa: N803
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        prevptr: Trigger = None,
        nextptr: Trigger = None,
        conditions: None = None,
        actions: None = None,
        *,
        preserved: Literal[True] = True,
        trigSection: bytes,  # noqa: N803
    ) -> None:
        ...

    def __init__(
        self,
        prevptr: Trigger = None,
        nextptr: Trigger = None,
        conditions: _Condition | None = None,
        actions: _Action | None = None,
        *,
        preserved: bool = True,
        trigSection: bytes | None = None,  # noqa: N803
    ) -> None:
        super().__init__()

        # Uses normal condition/action initialization
        if trigSection is None:
            # Normalize conditions/actions
            _conditions: Iterator[Condition]
            if conditions is None:
                _conditions = iter(())
            else:
                _conditions = ut.FlattenIter(conditions)

            _actions: Iterator[Action]
            if actions is None:
                _actions = iter(())
            else:
                _actions = ut.FlattenIter(actions)

            # Register condition/actions to trigger
            __conditions = []
            for i, cond in enumerate(_conditions):
                if i >= 16:
                    raise ut.EPError(_("Too many conditions"))
                if isinstance(cond, bool):
                    if cond:
                        cond = Condition(0, 0, 0, 0, 0, 22, 0, 0)  # Always
                    else:
                        cond = Condition(0, 0, 0, 0, 0, 23, 0, 0)  # Never
                cond.CheckArgs(i)
                __conditions.append(cond)

            __actions = []
            for i, act in enumerate(_actions):
                if i >= 64:
                    raise ut.EPError(_("Too many actions"))
                act.CheckArgs(i)
                __actions.append(act)

            for i, cond in enumerate(__conditions):
                cond.SetParentTrigger(self, i)
            for i, act in enumerate(__actions):
                act.SetParentTrigger(self, i)

            self._conditions = __conditions
            self._actions = __actions
            self._flags = 4 if preserved else 0

        # Uses trigger segment for initialization
        # NOTE : player information is lost inside eudplib.
        else:
            self._conditions, self._actions = [], []
            for i in range(16):
                condtype = trigSection[i * 20 + 15]
                if condtype == 22:
                    continue  # ignore Always, no worry disable/enable
                elif condtype >= 1:
                    condition = struct.unpack_from("<IIIHBBBBH", trigSection, i * 20)
                    self._conditions.append(
                        Condition(*condition[:8], eudx=condition[8])
                    )
            self._flags = trigSection[320 + 2048] & 5
            for i in range(64):
                acttype = trigSection[320 + i * 32 + 26]
                if acttype == 47:  # Comment
                    continue
                elif acttype >= 1:
                    action = struct.unpack_from(
                        "<IIIIIIHBBBBH", trigSection, 320 + i * 32
                    )
                    self._actions.append(Action(*action[:10], eudx=action[11]))

        if nextptr is not None and not IsConstExpr(nextptr):
            raise ut.EPError(_("nextptr should be ConstExpr"))

        # Register trigger to global table
        global _trg_counter
        _trg_counter += 1
        _register_trigger(self)  # This should be called before (1)

        # Set linked list pointers
        if prevptr is None:
            prevptr = 0
        if nextptr is None:
            nextptr = NextTrigger()  # (1)

        self._prevptr = prevptr
        self._nextptr = nextptr

    @property
    def preserved(self) -> bool:
        return bool(self._flags & 4)

    @preserved.setter
    def preserved(self, preserved: bool) -> None:
        if preserved:
            self._flags |= 4
        else:
            self._flags &= ~4

    def GetDataSize(self) -> int:  # noqa: N802
        return 2408

    def CollectDependency(self, pbuffer: ObjCollector) -> None:  # noqa: N802
        pbuffer.WriteDword(self._prevptr)
        pbuffer.WriteDword(self._nextptr)

        for cond in self._conditions:
            cond.CollectDependency(pbuffer)
        for act in self._actions:
            act.CollectDependency(pbuffer)

    # RawTrigger.WritePayload is overridden in Allocating and Writing phase
    # (See eudplib.core.allocator.payload.CreatePayload)
    def _allocate_trigger(self, pbuffer: alc.ObjAllocator) -> None:
        pbuffer._write_trigger(len(self._conditions), len(self._actions))

    def _write_trigger(self, pbuffer: alc.PayloadBuffer) -> None:
        pbuffer._write_trigger(
            self._prevptr,
            self._nextptr,
            # FIXME: Change Condition and Action to #[pyclass]
            (cond.fields for cond in self._conditions),
            (act.fields for act in self._actions),
            self._flags,
        )



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from collections.abc import Callable, Iterator
from typing import Literal, overload

from ..core import (
    Action,
    Add,
    AtLeast,
    ConstExpr,
    Deaths,
    DeathsX,
    EUDVariable,
    EUDXVariable,
    Exactly,
    Forward,
    IsEUDVariable,
    Memory,
    NextTrigger,
    PopTriggerScope,
    PushTriggerScope,
    RawTrigger,
    SeqCompute,
    SetDeathsX,
    SetMemory,
    SetNextPtr,
    SetNextTrigger,
    SetTo,
    Subtract,
    VProc,
)
from ..core.eudstruct.vararray import _InternalVArray
from ..ctrlstru import DoActions, EUDEndWhile, EUDSetContinuePoint
from ..ctrlstru.loopblock import _unsafe_whilenot
from ..localize import _
from ..memio.modcurpl import f_setcurpl2cpcache
from ..scdata import CurrentPlayer
from ..utils import EPD, EPError, EUDPeekBlock, ep_assert

_assign_helper = EUDVariable()


class UnitGroup:
    """Loop-efficient CUnit EPD collection

    ```js
    // epScript example

    // UnitGroup Declaration
    const zerglings = UnitGroup(1000);
    // max capacity = 1000, will use CPTrick

    // Register Unit
    zerglings.add(epd);

    // Loop UnitGroup
    foreach(unit : zerglings.cploop) {
        // Run Triggers on **any** zerglings (alive or dead)
        foreach(dead : unit.dying) {
            // Run Triggers on dead zerglings
        }  // <- dead zergling will be removed at end of *dying* block
        // Run Triggers on alive zerglings
    }

    // example usage
    function afterTriggerExec() {
        const zerglings = UnitGroup(1000);
        foreach(cunit : EUDLoopNewCUnit()) {
            epdswitch(cunit + 0x64/4, 255) {
            case $U("Zerg Zergling"):
                zerglings.add(epd);
                break;
            }
        }
        foreach(unit : zerglings.cploop) {
            foreach(dead : unit.dying) {
                // spawn Infested Terran when zergling dies
                dead.move_cp(0x4C / 4);  // Owner
                const owner = bread_cp(0, 0);
                dead.move_cp(0x28 / 4);  // Unit Position
                const x, y = posread_cp(0);

                setloc("loc", x, y);
                CreateUnit(1, "Infested Terran", "loc", owner);
            }
        }
    }
    ```
    """

    def __init__(self, capacity: int) -> None:
        self.capacity = capacity
        self.loopvar = EUDXVariable(EPD(0x6509B0), SetTo, 0)
        varray = _InternalVArray(capacity)(
            dest=self.loopvar, nextptr=self.loopvar.GetVTable()
        )
        self.varray = varray
        self.trg = EUDVariable(self.varray + 72 * capacity)
        self.pos = EUDVariable(EPD(self.varray) + 87 + 18 * capacity)
        self.cploop = _CPLoop(self)
        self._length = EUDVariable(0)

    @property  # FIXME: eudplib code can mutate length
    def length(self):
        return self._length

    def add(self, unit_epd) -> None:
        if IsEUDVariable(unit_epd):
            # Occupy 2T, Run 4T 10A
            SeqCompute(
                [
                    (self._length, Add, 1),
                    (self.trg, Add, -72),
                    (self.pos, Add, -18),
                    (unit_epd, Add, 0x4C // 4),
                    (EPD(unit_epd._varact) + 6, SetTo, 0x072D0000),
                    (EPD(unit_epd.getDestAddr()), None, self.pos),
                    (None, None, unit_epd),
                ]
            )
            SeqCompute([(unit_epd, Add, -(0x4C // 4))])
        else:
            # Occupy 1T, Run 4T 9A
            global _assign_helper
            SeqCompute(
                [
                    (self._length, Add, 1),
                    (self.trg, Add, -72),
                    (self.pos, Add, -18),
                    (_assign_helper, SetTo, unit_epd + 0x4C // 4),
                    (EPD(_assign_helper.getDestAddr()), None, self.pos),
                    (None, None, _assign_helper),
                ]
            )


class _CPLoop:
    def __init__(self, parent: UnitGroup) -> None:
        self._parent = parent

    def __iter__(self) -> Iterator[_CpHelper]:
        loopstart, pos, check_death = Forward(), Forward(), Forward()
        trg, tpos, loopvar, varray, capacity = (
            self._parent.trg,
            self._parent.pos,
            self._parent.loopvar,
            self._parent.varray,
            self._parent.capacity,
        )
        VProc(
            [trg, tpos],
            [
                trg.SetDest(EPD(loopstart)),
                tpos.SetDest(EPD(pos)),
                SetNextPtr(loopvar.GetVTable(), check_death),
            ],
        )

        def reset_cp() -> None:
            PushTriggerScope()
            reset_nextptr = RawTrigger(
                actions=SetNextPtr(loopvar.GetVTable(), check_death)
            )
            PopTriggerScope()
            nextptr = Forward()
            RawTrigger(
                nextptr=loopvar.GetVTable(),
                actions=[
                    SetNextPtr(loopvar.GetVTable(), reset_nextptr),
                    SetNextPtr(reset_nextptr, nextptr),
                ],
            )
            nextptr << NextTrigger()

        if _unsafe_whilenot()(Memory(loopstart, AtLeast, varray + 72 * capacity)):
            block = EUDPeekBlock("whileblock")[1]
            PushTriggerScope()  # remove entry
            remove_end = Forward()
            remove_start = RawTrigger(
                nextptr=trg.GetVTable(),
                actions=[
                    trg.SetDest(EPD(trg.GetVTable()) + 1),
                    SetNextPtr(loopvar.GetVTable(), remove_end),
                    loopvar.SetDest(0),  # pos가 이 액션의 값 칸
                ],
            )
            pos << remove_start + 328 + 32 * 2 + 20
            remove_end << RawTrigger(
                actions=[
                    self._parent._length.SubtractNumber(1),
                    trg.AddNumber(72),
                    tpos.AddNumber(18),
                    loopvar.SetDest(EPD(0x6509B0)),
                    SetNextPtr(loopvar.GetVTable(), check_death),
                ]
            )
            SetNextTrigger(block["contpoint"])
            PopTriggerScope()

            loopstart << block["loopstart"] + 4
            check_death << NextTrigger()

            def get_epd() -> EUDVariable:
                ret = EUDVariable()
                VProc(loopvar, loopvar.SetDest(ret))
                DoActions(
                    SetNextPtr(loopvar.GetVTable(), check_death),
                    loopvar.SetDest(EPD(0x6509B0)),
                    ret.SubtractNumber(19),
                )
                return ret

            # EUDIf()(DeathsX(CurrentPlayer, Exactly, 0, 0, 0xFF00)):
            yield _CpHelper(0x4C // 4, reset_cp, remove_start, get_epd)
            if not block["contpoint"].IsSet():
                EUDSetContinuePoint()
            DoActions(SetMemory(loopstart, Add, 72), SetMemory(pos, Add, 18))
        EUDEndWhile()

        f_setcurpl2cpcache()
        ep_assert(pos.IsSet(), "unit.dying must be added")


# TODO: Add CUnit-like convenient methods
class _CpHelper:
    def __init__(
        self,
        offset,
        resetf: Callable[[], None],
        remove: RawTrigger,
        epdf: Callable[[], EUDVariable],
    ) -> None:
        self.offset = offset
        self._reset = resetf
        self._remove_start = remove
        self.get_epd = epdf
        self.dying = _Dying(self)
        self._in_dying_block = False

    def remove(self) -> None:
        if self._in_dying_block:
            e = _("unit.remove() is automatically called at end of unit.dying block")
            raise EPError(e)
        SetNextTrigger(self._remove_start)

    @property  # read-only
    def epd(self) -> EUDVariable:
        return self.get_epd()

    @overload
    def set_cp(self, offset, *, action: Literal[False] = False) -> None:
        ...

    @overload
    def set_cp(self, offset, *, action: Literal[True]) -> Action:
        ...

    def set_cp(self, offset, *, action: bool = False) -> Action | None:
        self._reset()
        self.offset = 19
        return self.move_cp(offset, action=False)

    @overload
    def move_cp(self, offset, *, action: Literal[False] = False) -> None:
        ...

    @overload
    def move_cp(self, offset, *, action: Literal[True]) -> Action:
        ...

    def move_cp(self, offset, *, action: bool = False) -> Action | None:
        mod, val = None, None
        try:
            if offset > self.offset:
                mod, val = Add, offset - self.offset
            elif offset < self.offset:
                mod, val = Subtract, self.offset - offset
        except TypeError as e:
            if isinstance(offset, ConstExpr) and isinstance(self.offset, int):
                mod, val = Add, offset - self.offset
            elif isinstance(offset, int) and isinstance(self.offset, ConstExpr):
                mod, val = Subtract, self.offset - offset
            else:
                raise e
        self.offset = offset
        if mod and val:
            if action:
                return SetMemory(0x6509B0, mod, val)
            else:
                DoActions(SetMemory(0x6509B0, mod, val))
        return None


class _Dying:
    def __init__(self, helper: _CpHelper) -> None:
        self._helper: _CpHelper = helper

    def __iter__(self) -> Iterator[_CpHelper]:
        self._helper._in_dying_block = True
        dying_end, check_death, dying_block = Forward(), Forward(), Forward()
        RawTrigger(
            actions=[
                self._helper.move_cp(0x08 // 4, action=True),
                SetNextPtr(check_death, dying_end),
            ]
        )
        RawTrigger(
            conditions=Deaths(CurrentPlayer, Exactly, 0, 0),
            actions=[
                self._helper.move_cp(0x4C // 4, action=True),
                SetDeathsX(CurrentPlayer, SetTo, 0, 0, 0xFF00),
                self._helper.move_cp(0x08 // 4, action=True),
            ],
        )
        self._helper.move_cp(0x4C // 4)
        check_death << RawTrigger(
            nextptr=dying_end,
            conditions=DeathsX(CurrentPlayer, Exactly, 0, 0, 0xFF00),
            actions=SetNextPtr(check_death, dying_block),
        )
        dying_block << NextTrigger()
        yield self._helper
        self._helper._in_dying_block = False
        self._helper.remove()
        self._helper.offset = 19
        dying_end << NextTrigger()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .locf.locf import (
    f_addloc,
    f_dilateloc,
    f_getlocTL,
    f_setloc,
    f_setloc_epd,
)
from .mathf.atan2 import f_atan2, f_atan2_256
from .mathf.div import f_div_euclid, f_div_floor, f_div_towards_zero
from .mathf.lengthdir import f_lengthdir, f_lengthdir_256
from .mathf.pow import f_pow
from .mathf.sqrt import f_sqrt
from .utilf.binsearch import EUDBinaryMax, EUDBinaryMin
from .utilf.gametick import f_getgametick
from .utilf.listloop import (
    EUDLoopList,
    EUDLoopNewUnit,
    EUDLoopPlayerUnit,
    EUDLoopUnit,
    EUDLoopUnit2,
)
from .utilf.pexist import (
    EUDEndPlayerLoop,
    EUDLoopPlayer,
    EUDPlayerLoop,
    f_playerexist,
)
from .utilf.random import f_dwrand, f_getseed, f_rand, f_randomize, f_srand
from .utilf.userpl import (
    CenterViewAll,
    DisplayTextAll,
    IsUserCP,
    MinimapPingAll,
    PlayWAVAll,
    SetMissionObjectivesAll,
    TalkingPortraitAll,
    f_getuserplayerid,
)
from .wireframe.wireframe import (
    InitialWireframe,
    Is64BitWireframe,
    SetGrpWire,
    SetTranWire,
    SetWirefram,
    SetWireframes,
)

__all__ = [
    "f_addloc",
    "f_dilateloc",
    "f_getlocTL",
    "f_setloc",
    "f_setloc_epd",
    "f_atan2",
    "f_atan2_256",
    "f_div_euclid",
    "f_div_floor",
    "f_div_towards_zero",
    "f_lengthdir",
    "f_lengthdir_256",
    "f_pow",
    "f_sqrt",
    "EUDBinaryMax",
    "EUDBinaryMin",
    "f_getgametick",
    "EUDLoopList",
    "EUDLoopNewUnit",
    "EUDLoopPlayerUnit",
    "EUDLoopUnit",
    "EUDLoopUnit2",
    "EUDEndPlayerLoop",
    "EUDLoopPlayer",
    "EUDPlayerLoop",
    "f_playerexist",
    "f_dwrand",
    "f_getseed",
    "f_rand",
    "f_randomize",
    "f_srand",
    "CenterViewAll",
    "DisplayTextAll",
    "IsUserCP",
    "MinimapPingAll",
    "PlayWAVAll",
    "SetMissionObjectivesAll",
    "TalkingPortraitAll",
    "f_getuserplayerid",
    "InitialWireframe",
    "Is64BitWireframe",
    "SetGrpWire",
    "SetTranWire",
    "SetWirefram",
    "SetWireframes",
]



import random

from helper import *


@TestInstance
def test_parse():
    EP_SetRValueStrictMode(False)

    a = Db("dpdkfah")
    number, digit = f_parse(a)
    test_equality(
        "#0. Error handling: Not a number - return (0, 0)", [number, digit], [0, 0]
    )

    a = Db(" \t\n07285AC")
    number, digit = f_parse(a, 7)
    test_equality("#1. Parse 0 - return (0, 1)", [number, digit], [0, 1])
    number, digit = f_parse(a)
    test_equality("#2. Decimal number", [number, digit], [7285, 4])
    number, digit = f_parse(a, 16)
    test_equality("#3. Hexadecimal number", [number, digit], [0x7285AC, 6])

    a = Db("\f -0B1101bftx")
    number, digit = f_parse(a, 0)
    test_equality(
        "#4. radix=0: autodetect -Binary number",
        [number, digit],
        [-0b1101 & 0xFFFFFFFF, 4],
    )
    number, digit = f_parse(a, 16)
    test_equality(
        "#5. -Hexadecimal number", [number, digit], [-0xB1101BF & 0xFFFFFFFF, 7]
    )

    a = Db("  \r +0o7426 -f")
    number, digit = f_parse(a, 0)
    test_equality(
        "#6. radix=0: autodetect +Octal number", [number, digit], [0o7426, 4]
    )

    for n in range(7, 10):
        r = random.randint(0, 0x7FFFFFFF)
        a = Db(str(r))
        number, digit = f_parse(a)
        test_equality("#%u. Parse %u" % (n, r), [number, digit], [r, len(str(r))])

    a = Db("1" * 32)
    number, digit = f_parse(a)
    test_equality("Overflow: base 10", [number, digit], [0x7FFFFFFF, 10])
    for n in range(2, 37):
        # For overflow input, incorrect digits in base other than 10.
        number = f_parse(a, n)[0]
        test_equality("Overflow: base %u" % n, [number], [0x7FFFFFFF])

    EP_SetRValueStrictMode(True)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.eudfunc.eudf import _EUDPredefineParam, _EUDPredefineReturn
from . import cpmemio as cpm
from . import dwepdio as dwm
from . import modcurpl as cp
from . import readtable

cpcache = c.curpl.GetCPCache()

# Helper functions


def _lshift(a, b):
    if c.IsEUDVariable(a):
        return c.f_bitlshift(a, b)
    if isinstance(b, int) and b == 0:
        return a
    return a << b


@c.EUDFunc
def _wwriter(epd, subp, w):
    c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(3)):
        if cs.EUDSwitchCase()(i):
            c.f_bitlshift(w, 8 * i, ret=[w])
            cs.DoActions(c.SetDeathsX(cp.CP, c.SetTo, w, 0, 0xFFFF << (8 * i)))
            cs.EUDBreak()

    # Things gets complicated on this case.
    # We won't hand-optimize this case. This is a very, very rare case
    if cs.EUDSwitchCase()(3):
        b0, b1 = dwm.f_dwbreak(w)[2:4]
        cpm.f_bwrite_cp(0, 3, b0)
        cpm.f_bwrite_cp(1, 0, b1)

    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()


def f_wwrite_epd(epd, subp, w):
    if isinstance(subp, int) and isinstance(w, int):
        if subp == 3:
            return _wwriter(epd, subp, w)
        cs.DoActions(
            c.SetDeathsX(epd, c.SetTo, _lshift(w, 8 * subp), 0, 0xFFFF << (8 * subp))
        )
    else:
        _wwriter(epd, subp, w)


@c.EUDFunc
def _wadder(epd, subp, w):
    c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(3)):
        if cs.EUDSwitchCase()(i):
            c.f_bitlshift(w, 8 * i, ret=[w])
            cs.DoActions(c.SetDeathsX(cp.CP, c.Add, w, 0, 0xFFFF << (8 * i)))
            cs.EUDBreak()

    # Not implemented yet
    # if cs.EUDSwitchCase()(3):

    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()


def f_wadd_epd(epd, subp, w):
    if isinstance(subp, int) and isinstance(w, int):
        if subp == 3:
            return _wadder(epd, subp, w)
        cs.DoActions(
            c.SetDeathsX(epd, c.Add, _lshift(w, 8 * subp), 0, 0xFFFF << (8 * subp))
        )
    else:
        _wadder(epd, subp, w)


@c.EUDFunc
def _wsubtracter(epd, subp, w):
    c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(3)):
        if cs.EUDSwitchCase()(i):
            c.f_bitlshift(w, 8 * i, ret=[w])
            cs.DoActions(c.SetDeathsX(cp.CP, c.Subtract, w, 0, 0xFFFF << (8 * i)))
            cs.EUDBreak()

    # Not implemented yet
    # if cs.EUDSwitchCase()(3):

    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()


def f_wsubtract_epd(epd, subp, w):
    if isinstance(subp, int) and isinstance(w, int):
        if subp == 3:
            return _wsubtracter(epd, subp, w)
        cs.DoActions(
            c.SetDeathsX(
                epd, c.Subtract, _lshift(w, 8 * subp), 0, 0xFFFF << (8 * subp)
            )
        )
    else:
        _wsubtracter(epd, subp, w)


@c.EUDFunc
def _bwriter(epd, subp, b):
    c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(4)):
        if cs.EUDSwitchCase()(i):
            cs.DoActions(
                c.SetDeathsX(
                    cp.CP,
                    c.SetTo,
                    _lshift(b, 8 * i),
                    0,
                    0xFF << (8 * i),
                )
            )
            cs.EUDBreak()
    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()
    return b


def f_bwrite_epd(epd, subp, b) -> None:
    if isinstance(subp, int) and isinstance(b, int):
        cs.DoActions(
            c.SetDeathsX(epd, c.SetTo, _lshift(b, 8 * subp), 0, 0xFF << (8 * subp))
        )
    else:
        _bwriter(epd, subp, b)


def _bitwrite_epd(epd, subp, bit, b) -> None:
    ut.ep_assert(
        isinstance(bit, int) and bit.bit_count() == 1, f"bit_count is not 1: {bit}"
    )
    if isinstance(subp, int):
        bit <<= 8 * subp
        if c.IsEUDVariable(b):
            from ..trigger import Trigger

            cs.DoActions(c.SetDeathsX(epd, c.SetTo, 0, 0, bit))
            Trigger(
                conditions=b.AtLeast(1),
                actions=c.SetDeathsX(epd, c.SetTo, bit, 0, bit),
            )
        elif b == 0:
            cs.DoActions(c.SetDeathsX(epd, c.SetTo, 0, 0, bit))
        else:
            cs.DoActions(c.SetDeathsX(epd, c.SetTo, bit, 0, bit))
        return

    try:
        f = getattr(_bitwrite_epd, "f")
    except AttributeError:
        f = {}
        setattr(_bitwrite_epd, "f", f)

    try:
        _bitwriter = f[bit]
    except KeyError:

        @c.EUDFunc
        def _bitwriter(epd, subp, b):
            c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
            cs.EUDSwitch(subp)
            for i in ut._rand_lst(range(4)):
                if cs.EUDSwitchCase()(i):
                    c.RawTrigger(
                        actions=c.SetDeathsX(cp.CP, c.SetTo, 0, 0, bit << (8 * i))
                    )
                    c.RawTrigger(
                        conditions=b.AtLeast(1),
                        actions=c.SetDeathsX(
                            cp.CP, c.SetTo, bit << (8 * i), 0, bit << (8 * i)
                        ),
                    )
                    cs.EUDBreak()
            cs.EUDEndSwitch()
            cp.f_setcurpl2cpcache()
            return b

        f[bit] = _bitwriter

    return _bitwriter(epd, subp, b)


@c.EUDFunc
def _badder(epd, subp, b):
    c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(4)):
        if cs.EUDSwitchCase()(i):
            c.f_bitlshift(b, 8 * i, ret=[b])
            cs.DoActions(c.SetDeathsX(13, c.Add, b, 0, 0xFF << (8 * i)))
            cs.EUDBreak()
    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()
    return b


def f_badd_epd(epd, subp, b):
    if isinstance(subp, int) and isinstance(b, int):
        cs.DoActions(
            c.SetDeathsX(epd, c.Add, _lshift(b, 8 * subp), 0, 0xFF << (8 * subp))
        )
    else:
        _badder(epd, subp, b)


@c.EUDFunc
def _bsubtracter(epd, subp, b):
    c.VProc(epd, epd.SetDest(ut.EPD(0x6509B0)))
    cs.EUDSwitch(subp)
    for i in ut._rand_lst(range(4)):
        if cs.EUDSwitchCase()(i):
            c.f_bitlshift(b, 8 * i, ret=[b])
            cs.DoActions(c.SetDeathsX(cp.CP, c.Subtract, b, 0, 0xFF << (8 * i)))
            cs.EUDBreak()
    cs.EUDEndSwitch()
    cp.f_setcurpl2cpcache()
    return b


def f_bsubtract_epd(epd, subp, b):
    if isinstance(subp, int) and isinstance(b, int):
        cs.DoActions(
            c.SetDeathsX(
                epd, c.Subtract, _lshift(b, 8 * subp), 0, 0xFF << (8 * subp)
            )
        )
    else:
        _bsubtracter(epd, subp, b)


# -----------------------------


@_EUDPredefineReturn(1, 2)
@_EUDPredefineParam(cp.CP, 1)
@c.EUDFunc
def _wread_epd(epd, subp):
    w = _wread_epd._frets[0]
    fend = c.Forward()
    cs.EUDSwitch(subp)
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    block["_actions"].extend(
        [
            c.SetNextPtr(readtable.read_end_common, block["swend"]),
            c.SetMemory(readtable.copy_ret + 16, c.SetTo, ut.EPD(w.getValueAddr())),
            c.SetNextPtr(cpcache.GetVTable(), fend),
            cpcache.SetDest(ut.EPD(0x6509B0)),
        ]
    )
    for i in range(3):
        if cs.EUDSwitchCase()(i):
            readtrg = readtable._insert_or_get(0xFFFF << (8 * i), -(8 * i))
            c.SetNextTrigger(readtrg)

    # Things gets complicated on this case.
    # We won't hand-optimize this case. This is a very, very rare case
    if cs.EUDSwitchCase()(3):
        readtrg = readtable._insert_or_get(0xFF000000, -24)
        trg2, trg3 = c.Forward(), c.Forward()
        c.RawTrigger(
            nextptr=readtrg,
            actions=c.SetNextPtr(readtable.read_end_common, trg2),
        )
        readtrg = readtable._insert_or_get(0xFF, 8)
        trg2 << c.RawTrigger(
            nextptr=readtrg,
            actions=[
                c.SetMemory(0x6509B0, c.Add, 1),
                c.SetMemoryX(readtable.copy_ret + 24, c.SetTo, 8 << 24, 0xFF000000),
                c.SetNextPtr(readtable.read_end_common, trg3),
            ],
        )
        trg3 << c.RawTrigger(
            actions=c.SetMemoryX(
                readtable.copy_ret + 24, c.SetTo, 7 << 24, 0xFF000000
            )
        )

    cs.EUDEndSwitch()
    c.SetNextTrigger(cpcache.GetVTable())
    fend << c.NextTrigger()
    # return w


def f_wread_epd(epd, subp, *, ret=None):
    if isinstance(subp, int) and 0 <= subp <= 2:
        return readtable._epd_caller(
            readtable._insert_or_get(0xFFFF << (8 * subp), -(8 * subp))
        )(epd, ret=ret)

    return _wread_epd(epd, subp, ret=ret)


@_EUDPredefineReturn(1, 2)
@_EUDPredefineParam(cp.CP, 1)
@c.EUDFunc
def _bread_epd(epd, subp):
    b = _bread_epd._frets[0]
    fend = c.Forward()
    cs.EUDSwitch(subp)
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    block["_actions"].extend(
        [
            c.SetNextPtr(readtable.read_end_common, block["swend"]),
            c.SetMemory(readtable.copy_ret + 16, c.SetTo, ut.EPD(b.getValueAddr())),
            c.SetNextPtr(cpcache.GetVTable(), fend),
            cpcache.SetDest(ut.EPD(0x6509B0)),
        ]
    )
    for i in range(4):
        if cs.EUDSwitchCase()(i):
            readtrg = readtable._get(0xFF << (8 * i), -(8 * i))
            c.SetNextTrigger(readtrg)
    cs.EUDEndSwitch()
    c.SetNextTrigger(cpcache.GetVTable())
    fend << c.NextTrigger()
    # return b


def f_bread_epd(epd, subp, *, ret=None):
    if isinstance(subp, int) and 0 <= subp <= 3:
        return readtable._epd_caller(
            readtable._get(0xFF << (8 * subp), -(8 * subp))
        )(epd, ret=ret)

    return _bread_epd(epd, subp, ret=ret)


@_EUDPredefineReturn(1, 2)
@_EUDPredefineParam(cp.CP, 1)
@c.EUDFunc
def _boolread_epd(epd, subp):
    b = _boolread_epd._frets[0]
    fend = c.Forward()
    cs.EUDSwitch(subp)
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    block["_actions"].extend(
        [
            b.SetNumber(0),
            c.SetNextPtr(cpcache.GetVTable(), fend),
            cpcache.SetDest(ut.EPD(0x6509B0)),
        ]
    )
    for i in ut._rand_lst(range(4)):
        if cs.EUDSwitchCase()(i):
            c.RawTrigger(
                conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, 0xFF << (8 * i)),
                actions=b.SetNumber(1),
            )
            c.SetNextTrigger(cpcache.GetVTable())
    cs.EUDEndSwitch()
    fend << c.NextTrigger()
    # return b


def f_maskwrite_epd(epd, value, mask) -> None:
    cs.DoActions(c.SetDeathsX(epd, c.SetTo, value, 0, mask))


def _bitread_epd(bit):
    ut.ep_assert(
        isinstance(bit, int) and bit.bit_count() == 1, f"bit_count is not 1: {bit}"
    )
    f = getattr(_bitread_epd, "f", None)
    if f is None:
        f = {}
        _bitread_epd.f = f
    try:
        readfn = f[bit]
    except KeyError:

        @c.EUDFunc
        def readfn(epd, subp):
            b = c.EUDVariable()
            c.VProc(epd, [epd.SetDest(ut.EPD(0x6509B0)), b.SetNumber(0)])
            cs.EUDSwitch(subp)
            for i in ut._rand_lst(range(4)):
                if cs.EUDSwitchCase()(i):
                    mask = bit << (8 * i)
                    c.RawTrigger(
                        conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, mask),
                        actions=b.SetNumber(1),
                    )
                    cs.EUDBreak()
            cs.EUDEndSwitch()
            cp.f_setcurpl2cpcache()
            return b

        f[bit] = readfn

    return readfn



# Test for complex map

import os
import sys

sys.path.insert(0, os.path.abspath("..\\"))

from eudplib import *

"""
Creep reading
"""

mapwidth, mapheight, creepaddr_epd = EUDCreateVariables(3)


@EUDFunc
def f_creepread_init():
    # Get creepmap address
    creepaddr_epd << f_epdread_epd(EPD(0x6D0E84))
    SetVariables([mapwidth, mapheight], f_dwbreak(f_dwread_epd(EPD(0x57F1D4)))[0:2])


@EUDFunc
def f_creepread(x, y):
    ret = EUDCreateVariables(1)

    creepindex = mapwidth * y + x
    creepdwordindex, creepevenodd = f_div(creepindex, 2)

    # read tile data
    creepdat_epd = creepaddr_epd + creepdwordindex
    creepdat_word0, creepdat_word1 = f_dwbreak(f_dwread_epd(creepdat_epd))[0:2]

    # select word0/word1 by evenodd
    if EUDIf()(creepevenodd == 0):
        ret << creepdat_word0

    if EUDElse()():
        ret << creepdat_word1

    EUDEndIf()

    return ret


"""
Main logic
"""

LoadMap("outputmap/basemap/creeptest_basemap.scx")
# CompressPayload(True)


@EUDFunc
def main():
    # Iterate through each units.
    unitptr, unitepd = EUDCreateVariables(2)

    f_creepread_init()

    # Turbo rawtrigger

    if EUDWhile()(Always()):
        DoActions(SetDeaths(203151, SetTo, 1, 0))
        f_setcurpl(0)

        # Loop for every units
        for unitptr, unitepd in EUDLoopUnit():
            # check unittype
            # /*0x064*/ u16         unitType;
            unittype = f_dwbreak(f_dwread_epd(unitepd + (0x64 // 4)))[0]
            EUDContinueIfNot(unittype == EncodeUnit("Zerg Zergling"))

            # Get x, y coordinates of this unit.
            # uint16 unitx : unit + 0x28
            # uint16 unity : unit + 0x2A
            coord = f_dwread_epd(unitepd + (0x28 // 4))
            unitx, unity = f_dwbreak(coord)[0:2]

            # Convert coordinates to tile coordinates
            tileunitx = unitx // 32
            tileunity = unity // 32

            # creep -> continue
            creepval = f_creepread(tileunitx, tileunity)
            EUDContinueIf([creepval >= 16, creepval <= 31])

            # Slow down zergling.
            # Creating kakaru and killing them slows down zergling.
            SeqCompute(
                [
                    (EPD(0x58DC60 + 0), SetTo, unitx),
                    (EPD(0x58DC60 + 4), SetTo, unity),
                    (EPD(0x58DC60 + 8), SetTo, unitx),
                    (EPD(0x58DC60 + 12), SetTo, unity),
                ]
            )

            DoActions(
                CreateUnit(1, "Kakaru (Twilight Critter)", 1, Player1),
                RemoveUnitAt(All, "Kakaru (Twilight Critter)", 1, Player1),
            )

        DoActions(KillUnit("Kakaru (Twilight Critter)", Player1))

        EUDDoEvents()

    EUDEndWhile()


SaveMap("outputmap/testcreep.scx", main)



import random

from helper import *


@TestInstance
def test_arithmetics():
    EP_SetRValueStrictMode(False)

    v = random.randint(1, 0xFFFF)
    ev = EUDVariable()
    # const muldiv special cases
    for case in (-1, 0):
        ev << v
        ev *= case
        test_equality(f"{v} *= {case}", ev, (v * case) & 0xFFFFFFFF)
    for power in range(1, 8):
        case = 2**power
        ev << v
        ev *= case
        test_equality(f"{v} *= {case}", ev, (v * case) & 0xFFFFFFFF)

    ev << v
    ev //= 0
    test_equality(f"{v} //= 0", ev, 0xFFFFFFFF)
    ev << v
    ev %= 0
    test_equality(f"{v} %= 0", ev, v)
    ev << v
    ev //= 1
    test_equality(f"{v} //= 1", ev, v)
    ev << v
    ev %= 1
    test_equality(f"{v} %= 1", ev, 0)

    for _ in range(5):
        r = random.randint(3, 0xFFFFFFFF)
        ev << v
        ev *= r
        test_equality(f"{v} *= {r}", ev, (v * r) & 0xFFFFFFFF)

    for _ in range(5):
        r = random.randint(2, 0xFFFF)
        ev << v
        ev << f_pow(ev, r)
        test_equality(f"{v} ** {r}", ev, (v**r) & 0xFFFFFFFF)

    EP_SetRValueStrictMode(True)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import binascii
import os
import tempfile
from collections.abc import Callable

from ..bindings._rust import mpqapi
from ..core import RegisterCreatePayloadCallback
from ..core.eudfunc.trace.tracetool import _get_trace_map, _reset_trace_map
from ..core.mapdata import fixmapdata, mapdata
from ..localize import _
from ..utils import EPError
from .injector.apply_injector import apply_injector
from .injector.mainloop import main_starter
from .inlinecode.ilcprocesstrig import _preprocess_inline_code
from .mpqadd import _update_mpq

trace_header = None
trace_map = []


def get_trace_map() -> None:
    global trace_map, trace_header
    new_trace_header, new_trace_map = _get_trace_map()
    if not new_trace_header[1]:
        raise EPError(_("Empty trace header"))
    if new_trace_map:
        trace_header = new_trace_header
        trace_map = list(new_trace_map)
    _reset_trace_map()


RegisterCreatePayloadCallback(get_trace_map)


def SaveMap(fname: str, rootf: Callable, *, sector_size: int | None = None) -> None:  # noqa: N802
    """Save output map with root function.

    :param fname: Path for output map.
    :param rootf: Main entry function.
    """
    print(_("Saving to {}...").format(fname))
    chkt = mapdata.GetChkTokenized()

    _reset_trace_map()

    # Add payload
    root = main_starter(rootf)
    _preprocess_inline_code(chkt)
    mapdata.update_map_data()

    fixmapdata._fix_map_data(chkt)
    apply_injector(chkt, root)

    chkt.optimize()
    rawchk = chkt.savechk()
    print(f"\nOutput scenario.chk : {len(rawchk) / 1000000:.3f}MB")
    if sector_size is not None and isinstance(sector_size, int):
        from .loadmap import _load_map_path

        if _load_map_path is None:
            raise EPError(_("Must use LoadMap first"))

        # need to remove old file first because SFileCreateArchive2 does
        # not overwrite file and instead raise ERROR_ALREADY_EXISTS error
        if os.path.isfile(fname):
            try:
                os.remove(fname)
            except PermissionError as e:
                raise EPError(
                    _(
                        "You lack permission to access the output map"
                        "\n"
                        "Try turning off antivirus or StarCraft"
                    )
                ) from e

        try:
            mw = mpqapi.MPQ.clone_with_sector_size(
                _load_map_path, fname, sector_size
            )
        except Exception as e:
            raise EPError(_("Fail to access output map")) from e

        # add fake staredit\scenario.chk
        mw.set_file_locale(0)
        f = tempfile.NamedTemporaryFile(delete=False)
        f.write(b"")
        chk_path = f.name
        f.close()
        try:
            mw.add_file("staredit\\scenario.chk", chk_path)
        except Exception as e:
            raise EPError(_("Fail to add scenario.chk")) from e
        os.unlink(chk_path)
        mw.set_file_locale(0x409)

    else:
        # Process by modifying existing mpqfile
        try:
            with open(fname, "wb") as file:
                file.write(mapdata.GetRawFile())
        except PermissionError as e:
            raise EPError(
                _(
                    "You lack permission to access the output map"
                    "\n"
                    "Try turning off antivirus or StarCraft"
                )
            ) from e
        except Exception as e:
            raise EPError(_("Fail to access output map")) from e

        try:
            mw = mpqapi.MPQ.open(fname)
        except Exception as e:
            raise EPError(_("Fail to access output map")) from e

    # add real staredit\scenario.chk
    f = tempfile.NamedTemporaryFile(delete=False)
    f.write(bytes(rawchk))
    chk_path = f.name
    f.close()
    try:
        mw.add_file("staredit\\scenario.chk", chk_path)
    except Exception as e:
        raise EPError(_("Fail to add scenario.chk")) from e
    os.unlink(chk_path)
    mw.set_file_locale(0)

    _update_mpq(mw)
    try:
        mw.compact()
    except Exception as e:
        raise EPError(_("Fail to compact MPQ")) from e
    del mw

    if trace_map:
        trace_fname = fname + ".epmap"
        print(_("Writing trace file to {}").format(trace_fname))
        if trace_header is None:
            raise EPError(_("Unreachable callback error"))
        with open(trace_fname, "w", encoding="utf-8") as wf:
            wf.write(f"H0: {binascii.hexlify(trace_header[0]).decode('ascii')}\n")
            wf.write(f"H1: {binascii.hexlify(trace_header[1]).decode('ascii')}\n")
            for k, v in trace_map:
                wf.write(f" - {k:08}%08X : {v}\n")



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .allocator import (
    CompressPayload,
    ConstExpr,
    CreatePayload,
    Evaluate,
    Forward,
    GetObjectAddr,
    IsConstExpr,
    RegisterCreatePayloadCallback,
    RlocInt,
    RlocInt_C,
    ShufflePayload,
    toRlocInt,
)
from .calcf.bitwise import (
    f_bitand,
    f_bitlshift,
    f_bitnand,
    f_bitnor,
    f_bitnot,
    f_bitnxor,
    f_bitor,
    f_bitrshift,
    f_bitsplit,
    f_bitxor,
)
from .calcf.muldiv import f_div, f_mul
from .curpl import AddCurrentPlayer, SetCurrentPlayer
from .eudfunc.eudf import (
    EUDFunc,
    EUDTypedFunc,
)
from .eudfunc.eudfmethod import (
    EUDMethod,
    EUDTypedMethod,
)
from .eudfunc.eudfptr import EUDFuncPtr, EUDTypedFuncPtr
from .eudfunc.eudfuncn import EUDReturn
from .eudobj import Db, EUDObject
from .eudstruct import (
    EUDStruct,
    EUDVArray,
    selftype,
)
from .inlinens import (
    EUDClearNamespace,
    EUDRegistered,
    EUDRegisterObjectToNamespace,
    GetEUDNamespace,
)
from .mapdata import (
    TBL,
    GetChkTokenized,
    GetLocationIndex,
    GetPlayerInfo,
    GetPropertyIndex,
    GetStringIndex,
    GetSwitchIndex,
    GetUnitIndex,
    IsMapdataInitialized,
    UnitProperty,
)
from .rawtrigger import (
    Accumulate,
    Action,
    Add,
    All,
    AlliedVictory,
    Ally,
    Always,
    AtLeast,
    AtMost,
    Attack,
    Bring,
    Buildings,
    CenterView,
    Clear,
    Cleared,
    Command,
    CommandLeast,
    CommandLeastAt,
    CommandMost,
    CommandMostAt,
    Comment,
    Condition,
    CountdownTimer,
    CreateUnit,
    CreateUnitWithProperties,
    Custom,
    Deaths,
    DeathsX,
    Defeat,
    Disable,
    Disabled,
    DisplayText,
    Draw,
    ElapsedTime,
    Enable,
    EncodeAIScript,
    EncodeAllyStatus,
    EncodeButtonSet,
    EncodeComparison,
    EncodeCount,
    EncodeFlingy,
    EncodeIcon,
    EncodeImage,
    EncodeIscript,
    EncodeLocation,
    EncodeModifier,
    EncodeOrder,
    EncodePlayer,
    EncodePortrait,
    EncodeProperty,
    EncodePropState,
    EncodeResource,
    EncodeScore,
    EncodeSprite,
    EncodeString,
    EncodeSwitch,
    EncodeSwitchAction,
    EncodeSwitchState,
    EncodeTBL,
    EncodeTech,
    EncodeUnit,
    EncodeUnitOrder,
    EncodeUpgrade,
    EncodeWeapon,
    Enemy,
    Exactly,
    Gas,
    GetTriggerCounter,
    GiveUnits,
    HighestScore,
    Icon,
    Iscript,
    Kills,
    KillsAndRazings,
    KillUnit,
    KillUnitAt,
    LeaderBoardComputerPlayers,
    LeaderBoardControl,
    LeaderBoardControlAt,
    LeaderBoardGoalControl,
    LeaderBoardGoalControlAt,
    LeaderBoardGoalKills,
    LeaderBoardGoalResources,
    LeaderBoardGoalScore,
    LeaderBoardGreed,
    LeaderBoardKills,
    LeaderBoardResources,
    LeaderBoardScore,
    LeastKills,
    LeastResources,
    LowestScore,
    Memory,
    MemoryEPD,
    MemoryX,
    MemoryXEPD,
    MinimapPing,
    ModifyUnitEnergy,
    ModifyUnitHangarCount,
    ModifyUnitHitPoints,
    ModifyUnitResourceAmount,
    ModifyUnitShields,
    MostKills,
    MostResources,
    Move,
    MoveLocation,
    MoveUnit,
    MuteUnitSpeech,
    Never,
    NextTrigger,
    Opponents,
    Order,
    Ore,
    OreAndGas,
    Patrol,
    PauseGame,
    PauseTimer,
    PlayWAV,
    PopTriggerScope,
    Portrait,
    PreserveTrigger,
    PushTriggerScope,
    Random,
    RawTrigger,
    Razings,
    RemoveUnit,
    RemoveUnitAt,
    RunAIScript,
    RunAIScriptAt,
    Score,
    Set,
    SetAllianceStatus,
    SetCountdownTimer,
    SetDeaths,
    SetDeathsX,
    SetDoodadState,
    SetInvincibility,
    SetKills,
    SetMemory,
    SetMemoryEPD,
    SetMemoryX,
    SetMemoryXEPD,
    SetMissionObjectives,
    SetNextPtr,
    SetNextScenario,
    SetNextTrigger,
    SetResources,
    SetScore,
    SetSwitch,
    SetTo,
    StatText,
    Subtract,
    Switch,
    TalkingPortrait,
    Toggle,
    Total,
    Transmission,
    TrgAIScript,
    TrgAllyStatus,
    TrgComparison,
    TrgCount,
    TrgLocation,
    TrgModifier,
    TrgOrder,
    TrgProperty,
    TrgPropState,
    TrgResource,
    TrgScore,
    TrgString,
    TrgSwitch,
    TrgSwitchAction,
    TrgSwitchState,
    Units,
    UnitsAndBuildings,
    UnMuteUnitSpeech,
    UnpauseGame,
    UnpauseTimer,
    Victory,
    Wait,
)
from .variable import (
    EP_SetRValueStrictMode,
    EUDCreateVariables,
    EUDLightBool,
    EUDLightVariable,
    EUDVariable,
    EUDXVariable,
    IsEUDVariable,
    NonSeqCompute,
    SeqCompute,
    SetVariables,
    VProc,
)

__all__ = [
    "CompressPayload",
    "ConstExpr",
    "CreatePayload",
    "Evaluate",
    "Forward",
    "GetObjectAddr",
    "IsConstExpr",
    "RegisterCreatePayloadCallback",
    "RlocInt",
    "RlocInt_C",
    "ShufflePayload",
    "toRlocInt",
    "f_bitand",
    "f_bitlshift",
    "f_bitnand",
    "f_bitnor",
    "f_bitnot",
    "f_bitnxor",
    "f_bitor",
    "f_bitrshift",
    "f_bitsplit",
    "f_bitxor",
    "f_div",
    "f_mul",
    "AddCurrentPlayer",
    "SetCurrentPlayer",
    "EUDFunc",
    "EUDFuncPtr",
    "EUDMethod",
    "EUDReturn",
    "EUDTypedFunc",
    "EUDTypedFuncPtr",
    "EUDTypedMethod",
    "Db",
    "EUDObject",
    "EUDStruct",
    "EUDVArray",
    "selftype",
    "EUDClearNamespace",
    "EUDRegistered",
    "EUDRegisterObjectToNamespace",
    "GetEUDNamespace",
    "TBL",
    "GetChkTokenized",
    "GetLocationIndex",
    "GetPlayerInfo",
    "GetPropertyIndex",
    "GetStringIndex",
    "GetSwitchIndex",
    "GetUnitIndex",
    "IsMapdataInitialized",
    "UnitProperty",
    "Accumulate",
    "Action",
    "Add",
    "All",
    "AlliedVictory",
    "Ally",
    "Always",
    "AtLeast",
    "AtMost",
    "Attack",
    "Bring",
    "Buildings",
    "CenterView",
    "Clear",
    "Cleared",
    "Command",
    "CommandLeast",
    "CommandLeastAt",
    "CommandMost",
    "CommandMostAt",
    "Comment",
    "Condition",
    "CountdownTimer",
    "CreateUnit",
    "CreateUnitWithProperties",
    "Custom",
    "Deaths",
    "DeathsX",
    "Defeat",
    "Disable",
    "Disabled",
    "DisplayText",
    "Draw",
    "ElapsedTime",
    "Enable",
    "EncodeAIScript",
    "EncodeAllyStatus",
    "EncodeButtonSet",
    "EncodeComparison",
    "EncodeCount",
    "EncodeFlingy",
    "EncodeIcon",
    "EncodeImage",
    "EncodeIscript",
    "EncodeLocation",
    "EncodeModifier",
    "EncodeOrder",
    "EncodePlayer",
    "EncodePortrait",
    "EncodeProperty",
    "EncodePropState",
    "EncodeResource",
    "EncodeScore",
    "EncodeSprite",
    "EncodeString",
    "EncodeSwitch",
    "EncodeSwitchAction",
    "EncodeSwitchState",
    "EncodeTBL",
    "EncodeTech",
    "EncodeUnit",
    "EncodeUnitOrder",
    "EncodeUpgrade",
    "EncodeWeapon",
    "Enemy",
    "Exactly",
    "Gas",
    "GetTriggerCounter",
    "GiveUnits",
    "HighestScore",
    "Icon",
    "Iscript",
    "Kills",
    "KillsAndRazings",
    "KillUnit",
    "KillUnitAt",
    "LeaderBoardComputerPlayers",
    "LeaderBoardControl",
    "LeaderBoardControlAt",
    "LeaderBoardGoalControl",
    "LeaderBoardGoalControlAt",
    "LeaderBoardGoalKills",
    "LeaderBoardGoalResources",
    "LeaderBoardGoalScore",
    "LeaderBoardGreed",
    "LeaderBoardKills",
    "LeaderBoardResources",
    "LeaderBoardScore",
    "LeastKills",
    "LeastResources",
    "LowestScore",
    "Memory",
    "MemoryEPD",
    "MemoryX",
    "MemoryXEPD",
    "MinimapPing",
    "ModifyUnitEnergy",
    "ModifyUnitHangarCount",
    "ModifyUnitHitPoints",
    "ModifyUnitResourceAmount",
    "ModifyUnitShields",
    "MostKills",
    "MostResources",
    "Move",
    "MoveLocation",
    "MoveUnit",
    "MuteUnitSpeech",
    "Never",
    "NextTrigger",
    "Opponents",
    "Order",
    "Ore",
    "OreAndGas",
    "Patrol",
    "PauseGame",
    "PauseTimer",
    "PlayWAV",
    "PopTriggerScope",
    "Portrait",
    "PreserveTrigger",
    "PushTriggerScope",
    "Random",
    "RawTrigger",
    "Razings",
    "RemoveUnit",
    "RemoveUnitAt",
    "RunAIScript",
    "RunAIScriptAt",
    "Score",
    "Set",
    "SetAllianceStatus",
    "SetCountdownTimer",
    "SetDeaths",
    "SetDeathsX",
    "SetDoodadState",
    "SetInvincibility",
    "SetKills",
    "SetMemory",
    "SetMemoryEPD",
    "SetMemoryX",
    "SetMemoryXEPD",
    "SetMissionObjectives",
    "SetNextPtr",
    "SetNextScenario",
    "SetNextTrigger",
    "SetResources",
    "SetScore",
    "SetSwitch",
    "SetTo",
    "StatText",
    "Subtract",
    "Switch",
    "TalkingPortrait",
    "Toggle",
    "Total",
    "Transmission",
    "TrgAIScript",
    "TrgAllyStatus",
    "TrgComparison",
    "TrgCount",
    "TrgLocation",
    "TrgModifier",
    "TrgOrder",
    "TrgProperty",
    "TrgPropState",
    "TrgResource",
    "TrgScore",
    "TrgString",
    "TrgSwitch",
    "TrgSwitchAction",
    "TrgSwitchState",
    "Units",
    "UnitsAndBuildings",
    "UnMuteUnitSpeech",
    "UnpauseGame",
    "UnpauseTimer",
    "Victory",
    "Wait",
    "EP_SetRValueStrictMode",
    "EUDCreateVariables",
    "EUDLightBool",
    "EUDLightVariable",
    "EUDVariable",
    "EUDXVariable",
    "IsEUDVariable",
    "NonSeqCompute",
    "SeqCompute",
    "SetVariables",
    "VProc",
]



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) function switch_test(): EUDArray {
@EUDTypedFunc([], [EUDArray])
def f_switch_test():
    # (Line 2) const ret = EUDArray(6);
    ret = EUDArray(6)
    # (Line 3) static var x = 1234;
    x = _TYSV([None], [1234])
    # (Line 4) static var s = EPD(x.getValueAddr());
    s = _TYSV([None], [EPD(x.getValueAddr())])
    # (Line 6) switch(x) {}
    EUDSwitch(x)
    # (Line 7) epdswitch(s) {}
    EUDEndSwitch()
    EPDSwitch(s)
    # (Line 10) epdswitch (s, 240) {
    EUDEndSwitch()
    EPDSwitch(s, 240)
    # (Line 11) case 96:
    _t1 = EUDSwitchCase()
    # (Line 12) x += 1;
    if _t1(96):
        x.__iadd__(1)
        # (Line 13) break;
        EUDBreak()
        # (Line 14) case 160:
    _t2 = EUDSwitchCase()
    # (Line 15) x += 3;
    if _t2(160):
        x.__iadd__(3)
        # (Line 16) break;
        EUDBreak()
        # (Line 17) case 208:  // OK
    _t3 = EUDSwitchCase()
    # (Line 18) x += 5;
    if _t3(208):
        x.__iadd__(5)
        # (Line 19) break;
        EUDBreak()
        # (Line 20) default:
    # (Line 21) x = 0;
    if EUDSwitchDefault()():
        x << (0)
        # (Line 22) }
    # (Line 23) ret[0] = x;  // 1239
    EUDEndSwitch()
    _ARRW(ret, 0) << (x)
    # (Line 26) epdswitch (s, 56) {
    EPDSwitch(s, 56)
    # (Line 27) case 8:
    _t4 = EUDSwitchCase()
    # (Line 28) x += 2;
    if _t4(8):
        x.__iadd__(2)
        # (Line 29) break;
        EUDBreak()
        # (Line 30) case 16:  // OK
    _t5 = EUDSwitchCase()
    # (Line 31) x += 4;
    if _t5(16):
        x.__iadd__(4)
        # (Line 32) break;
        EUDBreak()
        # (Line 33) case 24:
    _t6 = EUDSwitchCase()
    # (Line 34) x += 6;
    if _t6(24):
        x.__iadd__(6)
        # (Line 35) break;
        EUDBreak()
        # (Line 36) case 32:
    _t7 = EUDSwitchCase()
    # (Line 37) x += 8;
    if _t7(32):
        x.__iadd__(8)
        # (Line 38) break;
        EUDBreak()
        # (Line 39) default:
    # (Line 40) x = 0;
    if EUDSwitchDefault()():
        x << (0)
        # (Line 41) }
    # (Line 42) ret[1] = x;  // 1243
    EUDEndSwitch()
    _ARRW(ret, 1) << (x)
    # (Line 45) switch (x, 432) {
    EUDSwitch(x, 432)
    # (Line 46) case 16:
    _t8 = EUDSwitchCase()
    # (Line 47) case 144:  // OK
    if _t8(16):
        pass
    _t9 = EUDSwitchCase()
    # (Line 48) x += 3;
    if _t9(144):
        x.__iadd__(3)
        # (Line 49) break;
        EUDBreak()
        # (Line 50) default:
    # (Line 51) x = 0;
    if EUDSwitchDefault()():
        x << (0)
        # (Line 52) }
    # (Line 53) ret[2] = x;  // 1246
    EUDEndSwitch()
    _ARRW(ret, 2) << (x)
    # (Line 56) switch (x, 211) {
    EUDSwitch(x, 211)
    # (Line 57) case 208:
    _t10 = EUDSwitchCase()
    # (Line 58) x += 5;
    if _t10(208):
        x.__iadd__(5)
        # (Line 59) break;
        EUDBreak()
        # (Line 60) case 210:
    _t11 = EUDSwitchCase()
    # (Line 61) x += 10;
    if _t11(210):
        x.__iadd__(10)
        # (Line 62) break;
        EUDBreak()
        # (Line 63) case 211:
    _t12 = EUDSwitchCase()
    # (Line 64) x += 15;
    if _t12(211):
        x.__iadd__(15)
        # (Line 65) break;
        EUDBreak()
        # (Line 66) default:
    # (Line 67) x = 0;
    if EUDSwitchDefault()():
        x << (0)
        # (Line 68) }
    # (Line 69) ret[3] = x;  // 1256
    EUDEndSwitch()
    _ARRW(ret, 3) << (x)
    # (Line 72) switch (x, 373) {
    EUDSwitch(x, 373)
    # (Line 73) case 96:  // OK
    _t13 = EUDSwitchCase()
    # (Line 74) x += 2;
    if _t13(96):
        x.__iadd__(2)
        # (Line 75) break;
        EUDBreak()
        # (Line 76) case 100:
    _t14 = EUDSwitchCase()
    # (Line 77) x += 5;
    if _t14(100):
        x.__iadd__(5)
        # (Line 78) break;
        EUDBreak()
        # (Line 79) case 112:
    _t15 = EUDSwitchCase()
    # (Line 80) x += 8;
    if _t15(112):
        x.__iadd__(8)
        # (Line 81) break;
        EUDBreak()
        # (Line 82) case 116:
    _t16 = EUDSwitchCase()
    # (Line 83) x += 11;
    if _t16(116):
        x.__iadd__(11)
        # (Line 84) break;
        EUDBreak()
        # (Line 85) default:
    # (Line 86) x = 0;
    if EUDSwitchDefault()():
        x << (0)
        # (Line 87) }
    # (Line 88) ret[4] = x;  // 1258
    EUDEndSwitch()
    _ARRW(ret, 4) << (x)
    # (Line 91) switch (x, 511) {
    EUDSwitch(x, 511)
    # (Line 92) case 121:
    _t17 = EUDSwitchCase()
    # (Line 93) x += 1;
    if _t17(121):
        x.__iadd__(1)
        # (Line 94) break;
        EUDBreak()
        # (Line 95) case 179:
    _t18 = EUDSwitchCase()
    # (Line 96) x += 4;
    if _t18(179):
        x.__iadd__(4)
        # (Line 97) break;
        EUDBreak()
        # (Line 98) case 234:  // OK
    _t19 = EUDSwitchCase()
    # (Line 99) x += 2;
    if _t19(234):
        x.__iadd__(2)
        # (Line 100) break;
        EUDBreak()
        # (Line 101) case 338:
    _t20 = EUDSwitchCase()
    # (Line 102) x += 8;
    if _t20(338):
        x.__iadd__(8)
        # (Line 103) break;
        EUDBreak()
        # (Line 104) case 428:
    _t21 = EUDSwitchCase()
    # (Line 105) x += 5;
    if _t21(428):
        x.__iadd__(5)
        # (Line 106) break;
        EUDBreak()
        # (Line 107) case 453:
    _t22 = EUDSwitchCase()
    # (Line 108) x += 7;
    if _t22(453):
        x.__iadd__(7)
        # (Line 109) break;
        EUDBreak()
        # (Line 110) default:
    # (Line 111) x = 0;
    if EUDSwitchDefault()():
        x << (0)
        # (Line 112) }
    # (Line 113) ret[5] = x;  // 1260
    EUDEndSwitch()
    _ARRW(ret, 5) << (x)
    # (Line 114) return ret;
    EUDReturn(ret)
    # (Line 115) }
    # (Line 117) const ack = PVariable(list(1, 2, 3, 4, 5, 6, 7, 8));

ack = _CGFW(lambda: [PVariable(FlattenList([1, 2, 3, 4, 5, 6, 7, 8]))], 1)[0]
# (Line 118) const ackMax = 8;
ackMax = _CGFW(lambda: [8], 1)[0]
# (Line 119) function test_array() {
@EUDFunc
def f_test_array():
    # (Line 120) var ret = 0;
    ret = _TYLV([None], [0])
    # (Line 121) const p = EUDVariable(1);
    p = EUDVariable(1)
    # (Line 122) ret += ack[p] % ackMax > 4 ? 1 : 2;  // ret = 2
    ret.__iadd__(EUDTernary(ack[p] % ackMax <= 4, neg=True)(1)(2))
    # (Line 124) const arr = [1, 2, 3, 4, 5, 6, 7, 8];
    arr = _ARR(FlattenList([1, 2, 3, 4, 5, 6, 7, 8]))
    # (Line 125) ack[0] %= arr[p];
    _ARRW(ack, 0).__imod__(arr[p])
    # (Line 126) ret = ret >> 0;  // ret = 2
    ret << (ret >> 0)
    # (Line 128) ack[p] &= 1;  // ack[1] = 0
    _ARRW(ack, p).__iand__(1)
    # (Line 129) ack[p] &= arr[p];
    _ARRW(ack, p).__iand__(arr[p])
    # (Line 130) ack[p] -= -2;  // ack[1] = 2
    _ARRW(ack, p).__isub__(-2)
    # (Line 131) const x = ack[p];
    x = ack[p]
    # (Line 132) ack[p] = x & arr[p];  // 2 & 2
    _ARRW(ack, p) << (x & arr[p])
    # (Line 133) arr[p] &= ack[p];
    _ARRW(arr, p).__iand__(ack[p])
    # (Line 134) ret *= ack[p] * arr[p];  // ret = 8
    ret.__imul__(ack[p] * arr[p])
    # (Line 136) arr[p] ^= ack[p];  // arr[1] = 0
    _ARRW(arr, p).__ixor__(ack[p])
    # (Line 137) arr[p] |= ack[0];  // arr[1] = 1
    _ARRW(arr, p).__ior__(ack[0])
    # (Line 138) arr[p] <<= ack[p];  // arr[1] = 4
    _ARRW(arr, p).__ilshift__(ack[p])
    # (Line 139) ret |= arr[p];  // ret = 12
    ret.__ior__(arr[p])
    # (Line 141) if (ack[p] > arr[p]) ret = 0;
    if EUDIf()(_ARRC(ack, p) <= arr[p], neg=True):
        ret << (0)
        # (Line 142) if (arr[p]) ret <<= 1;  // ret = 24
    EUDEndIf()
    if EUDIf()(arr[p]):
        ret.__ilshift__(1)
        # (Line 143) ret -= ack[p] + p / 2;  // ret = 22
    EUDEndIf()
    ret.__isub__(ack[p] + p // 2)
    # (Line 145) switch(arr[p]) {
    EUDSwitch(arr[p])
    # (Line 146) case 4:  // fall-through
    _t3 = EUDSwitchCase()
    # (Line 147) case 1:
    if _t3(4):
        pass
    _t4 = EUDSwitchCase()
    # (Line 148) ret += 1;  // ret = 23
    if _t4(1):
        ret.__iadd__(1)
        # (Line 149) }
    # (Line 150) switch(ack[p]) {
    EUDEndSwitch()
    EUDSwitch(ack[p])
    # (Line 151) case 2:
    _t5 = EUDSwitchCase()
    # (Line 152) break;
    if _t5(2):
        EUDBreak()
        # (Line 153) default:
    # (Line 154) ret = 0;
    if EUDSwitchDefault()():
        ret << (0)
        # (Line 155) }
    # (Line 157) return ret;  // ret = 23
    EUDEndSwitch()
    EUDReturn(ret)
    # (Line 158) }
    # (Line 160) function test_all_actions() {

@EUDFunc
def f_test_all_actions():
    # (Line 161) var sound = $T("sound\\Misc\\PError.WAV");
    sound = _TYLV([None], [GetStringIndex("sound\\Misc\\PError.WAV")])
    # (Line 162) PlayWAVAll(sound);
    # (Line 163) }
    _ALL(PlayWAVAll(sound))



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable
from typing import Literal, overload

from ... import core as c
from ... import utils as ut
from ...core.eudfunc.eudf import _EUDPredefineParam
from ...localize import _
from ...memio import f_dwread_cp, f_posread_cp, f_setcurpl2cpcache
from ...memio import modcurpl as cp

_loct = ut.EPD(0x58DC60) - 5


def _locfgen2(mod1, mod2, mod3, mod4, signed=False) -> Callable:
    @c.EUDFunc
    def _locf(epd, x, y) -> None:
        act = c.Forward()

        c.VProc(
            [epd, x],
            [
                epd.AddNumber(_loct),
                epd.SetDest(ut.EPD(act) + 4),
                x.SetDest(ut.EPD(act) + 5)
                if not signed
                else [
                    c.SetMemory(act + 20, c.SetTo, ~0),
                    x.QueueSubtractTo(ut.EPD(act) + 5),
                ],
            ],
        )
        settb = c.Forward()
        c.RawTrigger(
            nextptr=epd.GetVTable(),
            actions=[
                c.SetMemory(act + 20, c.Add, 1) if signed else [],
                epd.AddNumber(2),
                epd.AddDest(16),
                x.AddDest(16),
                c.SetNextPtr(x.GetVTable(), settb),
                c.SetMemory(x._varact + 24, c.Subtract, 0x02000000)
                if signed
                else [],
            ],
        )
        settb << c.NextTrigger()
        c.VProc(
            [epd, y],
            [
                epd.AddNumber(-1),
                epd.AddDest(-8),
                y.SetDest(ut.EPD(act) + 8 + 5)
                if not signed
                else [
                    c.SetMemory(act + 52, c.SetTo, ~0),
                    y.QueueSubtractTo(ut.EPD(act) + 8 + 5),
                ],
            ],
        )
        setcoords = c.Forward()
        c.RawTrigger(
            nextptr=epd.GetVTable(),
            actions=[
                c.SetMemory(act + 52, c.Add, 1) if signed else [],
                epd.AddNumber(2),
                epd.AddDest(16),
                y.AddDest(16),
                c.SetNextPtr(y.GetVTable(), setcoords),
                c.SetMemory(y._varact + 24, c.Subtract, 0x02000000)
                if signed
                else [],
            ],
        )
        setcoords << c.NextTrigger()
        c.RawTrigger(
            actions=[
                act << c.SetDeaths(0, mod1, 0, 0),
                c.SetDeaths(0, mod2, 0, 0),
                c.SetDeaths(0, mod3, 0, 0),
                c.SetDeaths(0, mod4, 0, 0),
            ]
        )

    return _locf


def _locfgen4(mod1, mod2, mod3, mod4, signed=False) -> Callable:
    @c.EUDFunc
    def _locf(epd, left, t, r, b) -> None:
        act = c.Forward()

        c.VProc(
            [epd, left],
            [
                epd.AddNumber(_loct),
                epd.SetDest(ut.EPD(act) + 4),
                left.SetDest(ut.EPD(act) + 5)
                if not signed
                else [
                    c.SetMemory(act + 20, c.SetTo, ~0),
                    left.QueueSubtractTo(ut.EPD(act) + 5),
                ],
            ],
        )
        c.VProc(
            [epd, t],
            [
                c.SetMemory(act + 20, c.Add, 1) if signed else [],
                epd.AddNumber(1),
                epd.AddDest(8),
                t.SetDest(ut.EPD(act) + 8 + 5)
                if not signed
                else [
                    c.SetMemory(act + 52, c.SetTo, ~0),
                    t.QueueSubtractTo(ut.EPD(act) + 8 + 5),
                ],
            ],
        )
        c.VProc(
            [epd, r],
            [
                c.SetMemory(act + 52, c.Add, 1) if signed else [],
                epd.AddNumber(1),
                epd.AddDest(8),
                r.SetDest(ut.EPD(act) + 16 + 5),
            ],
        )
        c.VProc(
            [epd, b],
            [
                epd.AddNumber(1),
                epd.AddDest(8),
                b.SetDest(ut.EPD(act) + 24 + 5),
            ],
        )
        c.RawTrigger(
            actions=[
                act << c.SetDeaths(0, mod1, 0, 0),
                c.SetDeaths(0, mod2, 0, 0),
                c.SetDeaths(0, mod3, 0, 0),
                c.SetDeaths(0, mod4, 0, 0),
            ]
        )

    return _locf


_SetLoc2 = _locfgen2(c.SetTo, c.SetTo, c.SetTo, c.SetTo)
_AddLoc2 = _locfgen2(c.Add, c.Add, c.Add, c.Add)
_DilateLoc2 = _locfgen2(c.Add, c.Add, c.Add, c.Add, signed=True)
_SetLoc4 = _locfgen4(c.SetTo, c.SetTo, c.SetTo, c.SetTo)
_AddLoc4 = _locfgen4(c.Add, c.Add, c.Add, c.Add)
_DilateLoc4 = _locfgen4(c.Add, c.Add, c.Add, c.Add, signed=True)


@overload
def f_setloc(loc, *coords, action: Literal[True]) -> list[c.Action]:
    ...


@overload
def f_setloc(loc, *coords, action: Literal[False]) -> None:
    ...


def f_setloc(loc, *coords, action=False):
    ut.ep_assert(
        len(coords) == 2 or len(coords) == 4,
        _("number of coordinates should be 2 or 4."),
    )
    loc = c.EncodeLocation(loc)
    if action is True:
        ut.ep_assert(all(c.IsConstExpr(x) for x in coords) and c.IsConstExpr(loc))
    if c.IsConstExpr(loc):
        dst = _loct + 5 * loc
        if len(coords) == 2:
            left, t = coords
            r, b = coords
        else:
            left, t, r, b = coords
        if action is True:
            return [
                c.SetMemoryEPD(dst + i, c.SetTo, x)
                for i, x in enumerate((left, t, r, b))
            ]
        c.NonSeqCompute(
            [
                (dst, c.SetTo, left),
                (dst + 1, c.SetTo, t),
                (dst + 2, c.SetTo, r),
                (dst + 3, c.SetTo, b),
            ]
        )
    elif len(coords) == 2:
        _SetLoc2(loc * 5, *coords)
    else:
        _SetLoc4(loc * 5, *coords)


@overload
def f_addloc(loc, *coords, action: Literal[True]) -> list[c.Action]:
    ...


@overload
def f_addloc(loc, *coords, action: Literal[False]) -> None:
    ...


def f_addloc(loc, *coords, action=False):
    ut.ep_assert(
        len(coords) == 2 or len(coords) == 4,
        _("number of coordinates should be 2 or 4."),
    )
    loc = c.EncodeLocation(loc)
    if action is True:
        ut.ep_assert(all(c.IsConstExpr(x) for x in coords) and c.IsConstExpr(loc))
    if c.IsConstExpr(loc):
        dst = _loct + 5 * loc
        if len(coords) == 2:
            left, t = coords
            r, b = coords
        else:
            left, t, r, b = coords
        if action is True:
            return [
                c.SetMemoryEPD(dst + i, c.Add, x)
                for i, x in enumerate((left, t, r, b))
                if not (isinstance(x, int) and x == 0)
            ]
        c.NonSeqCompute(
            [
                (dst + i, c.Add, x)
                for i, x in enumerate((left, t, r, b))
                if not (isinstance(x, int) and x == 0)
            ]
        )
    elif len(coords) == 2:
        _AddLoc2(loc * 5, *coords)
    else:
        _AddLoc4(loc * 5, *coords)


@overload
def f_dilateloc(loc, *coords, action: Literal[True]) -> list[c.Action]:
    ...


@overload
def f_dilateloc(loc, *coords, action: Literal[False]) -> None:
    ...


def f_dilateloc(loc, *coords, action=False):
    ut.ep_assert(
        len(coords) == 2 or len(coords) == 4,
        _("number of coordinates should be 2 or 4."),
    )
    loc = c.EncodeLocation(loc)
    if action is True:
        ut.ep_assert(all(c.IsConstExpr(x) for x in coords) and c.IsConstExpr(loc))
    if c.IsConstExpr(loc):
        dst = _loct + 5 * loc
        if len(coords) == 2:
            left, t = coords
            r, b = coords
        else:
            left, t, r, b = coords
        if action is True:
            return [
                c.SetMemoryEPD(dst + i, c.Add, x)
                for i, x in enumerate((-left, -t, r, b))
                if not (isinstance(x, int) and x == 0)
            ]
        c.NonSeqCompute(
            [
                (dst + i, c.Add, x)
                for i, x in enumerate((-left, -t, r, b))
                if not (isinstance(x, int) and x == 0)
            ]
        )
    elif len(coords) == 2:
        _DilateLoc2(loc * 5, *coords)
    else:
        _DilateLoc4(loc * 5, *coords)


@c.EUDFunc
def _getloc_tl(epd):
    c.VProc(epd, [epd.AddNumber(_loct), epd.SetDest(ut.EPD(0x6509B0))])
    left = f_dwread_cp(0)
    c.RawTrigger(actions=c.SetMemory(0x6509B0, c.Add, 1))
    f_dwread_cp(0, ret=[epd])
    f_setcurpl2cpcache()
    return left, epd


def f_getlocTL(loc, **kwargs) -> tuple[c.EUDVariable, c.EUDVariable]:  # noqa: N802
    """
    로케이션의 위(top), 왼쪽 (left) 좌표를 얻어냅니다.
    @param {[type]} loc 로케이션 번호. $L(로케이션 이름) 으로 얻을 수 있습니다.
    """
    loc = c.EncodeLocation(loc)
    return _getloc_tl(loc * 5, **kwargs)


_set_loc: c.Action = c.SetMemory(0x6509B0, c.SetTo, 0)
_setcp2loc: c.ConstExpr = ut.EPD(_set_loc) + 5


@_EUDPredefineParam((_setcp2loc,), cp.CP)
@c.EUDFunc
def _setloc_epd(loc, epd):
    global _setcp2loc
    set_x_epd = c.Forward()
    f_posread_cp(0, ret=[set_x_epd, set_x_epd + 16])

    c.RawTrigger(
        actions=[
            _set_loc,
            c.SetMemory(0x6509B0, c.Add, _loct),
        ]
    )

    set_x = c.SetDeaths(cp.CP, c.SetTo, 0, 0)
    set_xy = c.RawTrigger(
        actions=[
            set_x,
            c.SetMemory(0x6509B0, c.Add, 1),
            c.SetDeaths(cp.CP, c.SetTo, 0, 0),
        ]
    )
    set_x_epd << ut.EPD(set_x) + 5

    done = c.Forward()
    one_more = c.RawTrigger(
        nextptr=set_xy,
        actions=[
            c.SetMemory(0x6509B0, c.Add, 1),
            c.SetNextPtr(set_xy, done),
        ],
    )
    done << f_setcurpl2cpcache(actions=c.SetNextPtr(set_xy, one_more))


def f_setloc_epd(loc, epd) -> None:
    loc = c.EncodeLocation(loc)
    _setloc_epd(loc * 5, epd)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .eudstruct import EUDStruct
from .selftype import selftype
from .vararray import EUDVArray

__all__ = [
    "EUDStruct",
    "selftype",
    "EUDVArray",
]



# Copyright 2024 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

class MPQ:
    @staticmethod
    def open(path: str) -> MPQ: ...
    @staticmethod
    def create(path: str, sector_size: int = 3, file_count: int = 1024) -> MPQ: ...
    def get_file_names_from_listfile(self) -> list[str]: ...
    def extract_file(self, file_path: str) -> bytes: ...
    def add_file(
        self, archived_name: str, file_path: str, replace_existing: bool = True
    ) -> None: ...
    @staticmethod
    def set_file_locale(file_locale: int) -> None: ...
    def get_max_file_count(self) -> int: ...
    def set_max_file_count(self, count: int) -> None: ...
    def compact(self) -> None: ...
    @staticmethod
    def clone_with_sector_size(
        input_path: str, output_path: str, sector_size: int
    ) -> MPQ: ...



#!/usr/bin/python
# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import os
import re
import sys
import types
from bisect import bisect_right
from importlib.machinery import FileFinder, SourceFileLoader

from ..localize import _
from ..utils import EPError
from .epscompile import epsCompile
from .linetable_calculator import (
    PYCODE_ATTRIBUTES,
    gen_code_options,
    gen_new_opcode,
)

lineno_regex = re.compile(b" *# \\(Line (\\d+)\\) (.+)")
is_scdb_map = False


def IsSCDBMap():  # noqa: N802
    return is_scdb_map


def _modify_code_linetable(codeobj: types.CodeType, ep_lineno_map):
    # See: https://github.com/python/cpython/blob/main/Objects/locations.md
    # https://github.com/python/cpython/blob/c1652d6d6201e5407b94afc297115a584b5a0955/Python/assemble.c#L231-L242
    code_options = gen_code_options(codeobj)

    # For code objects
    new_co_consts = []
    for c in code_options["co_consts"]:
        if isinstance(c, types.CodeType):
            c = _modify_code_linetable(c, ep_lineno_map)
        new_co_consts.append(c)

    code_options["co_consts"] = new_co_consts

    return gen_new_opcode(codeobj, code_options, PYCODE_ATTRIBUTES, ep_lineno_map)


# TODO: refactor code or remove as duplicate
def _modify_code_lnotab(codeobj: types.CodeType, ep_lineno_map):
    # See: https://peps.python.org/pep-0626/#backwards-compatibility
    # https://github.com/python/cpython/blob/main/Objects/lnotab_notes.txt
    co_lnotab = codeobj.co_lnotab
    co_firstlineno = codeobj.co_firstlineno

    # Reconstruct co_lnotab
    new_lnotab = []
    current_line = co_firstlineno
    current_mapped_line = ep_lineno_map(current_line)
    for i in range(0, len(co_lnotab), 2):
        bytecode_len, line_advance = co_lnotab[i : i + 2]
        next_line = current_line + line_advance
        next_mapped_line = ep_lineno_map(next_line)
        new_line_advance = next_mapped_line - current_mapped_line
        while new_line_advance >= 0xFF:
            new_lnotab.append(bytes([0, 0xFF]))
            new_line_advance -= 0xFF
        new_lnotab.append(bytes([bytecode_len, new_line_advance]))
        current_line = next_line
        current_mapped_line = next_mapped_line

    # For code objects
    new_co_consts = []
    for c in codeobj.co_consts:
        if isinstance(c, types.CodeType):
            c = _modify_code_lnotab(c, ep_lineno_map)
        new_co_consts.append(c)

    # Python 3.7~3.10
    codeobj = types.CodeType(
        codeobj.co_argcount,
        codeobj.co_posonlyargcount,  # python 3.8 support (See PEP 570)
        codeobj.co_kwonlyargcount,
        codeobj.co_nlocals,
        codeobj.co_stacksize,
        codeobj.co_flags,
        codeobj.co_code,
        tuple(new_co_consts),
        codeobj.co_names,
        codeobj.co_varnames,
        codeobj.co_filename,
        codeobj.co_name,
        ep_lineno_map(co_firstlineno),  # codeobj.co_firstlineno,
        b"".join(new_lnotab),  # type: ignore
        codeobj.co_freevars,  # type: ignore
        codeobj.co_cellvars,  # type: ignore
    )

    return codeobj


class EPSLoader(SourceFileLoader):
    def create_module(self, spec):
        module_name = spec.name
        module = types.ModuleType(module_name)
        module.__name__ = module_name
        module.__loader__ = self
        sys.modules[module_name] = module
        return module

    def get_data(self, path):
        """Return the data from path as raw bytes."""
        global is_scdb_map
        with open(path, "rb") as file:
            file_data = file.read()
            if path.endswith(".pyc") or path.endswith(".pyo"):
                return file_data
            if "SCDB.eps" in os.path.relpath(path):
                is_scdb_map = True
            print(_('[epScript] Compiling "{}"...').format(os.path.relpath(path)))
            compiled = epsCompile(path, file_data)
        if compiled is None:
            raise EPError(_(" - Compiled failed for {}").format(path))
        dirname, filename = os.path.split(path)
        epsdir = os.path.join(dirname, "__epspy__")
        try:
            if not os.path.isdir(epsdir):
                os.mkdir(epsdir)
            ofname = os.path.splitext(filename)[0] + ".py"
            ofname = os.path.join(epsdir, ofname)
            with open(ofname, "w", encoding="utf-8") as file:
                file.write(compiled.decode("utf-8"))
        except OSError:
            pass

        return compiled

    def source_to_code(self, data, path, *, _optimize=-1):
        codeobj = super().source_to_code(data, path, _optimize=_optimize)

        # Read lines from code data
        code_line = [0]
        ep_lineno_map = [0]
        data = data.replace(b"\r\n", b"\n")
        for lineno, line in enumerate(data.split(b"\n")):
            match = lineno_regex.match(line)
            if match:
                code_line.append(lineno + 1)
                ep_lineno_map.append(int(match.group(1)))

        # Reconstruct code data
        def line_mapper(line):
            # FIXME: sometimes return lineno off-by-one
            lineno = ep_lineno_map[bisect_right(code_line, line) - 1]
            if line != 0:
                lineno = max(0, lineno)
            return lineno

        if sys.version_info >= (3, 11):
            codeobj = _modify_code_linetable(codeobj, line_mapper)
        else:
            codeobj = _modify_code_lnotab(codeobj, line_mapper)
        return codeobj


class EPSFinder:
    def __init__(self):
        self._finderCache = {}

    def _get_finder(self, path):
        try:
            return self._finderCache[path]
        except KeyError:
            self._finderCache[path] = FileFinder(path, (EPSLoader, [".eps"]))
            return self._finderCache[path]

    def find_spec(self, fullname, path, target=None):
        if path is None:
            path = sys.path
        for path_entry in path:
            finder = self._get_finder(path_entry)
            spec = finder.find_spec(fullname)
            if spec is None:
                continue
            return spec


sys.meta_path.append(EPSFinder())
